import json
import os
import time

import io
import hashlib
import base64

from typing import Any


class IO():
    """
    This is for normal IO processing
    """

    def __init__(self):
        self.current_dir: str = os.getcwd()
        self.__log_path: str = os.path.join(self.current_dir, '.log')

    def make_sure_sudo_permission(self):
        """
        exit if you don't have sudo permission
        """
        if os.getuid() != 0:
            print("\n I only got my super power if you run me with sudo!")
            exit()

    def read(self, file_path: str, auto_detect_encoding=False) -> str:
        """
        read text from txt file

        Parameters
        ----------
        file_path
            txt file path
        """
        encoding = 'utf-8'

        if (auto_detect_encoding == True):
            import chardet
            rawdata = open(file_path, "rb").read()
            result = chardet.detect(rawdata)
            if (result != None):
                encoding = result['encoding']

        if os.path.exists(file_path):
            with open(file_path, 'r', encoding=encoding, errors="ignore") as f:
                result = f.read()
            return result
        else:
            print(f"File '{file_path}' does not exists.")
            return ""

    def write(self, file_path: str, content: str):
        """
        write text into txt file

        Parameters
        ----------
        file_path
            target txt file path
        content
            text string
        """
        with open(file_path, 'w', encoding="utf-8", errors="ignore") as f:
            f.write(content)

    def append(self, file_path: str, content: str):
        """
        append text at the end of a txt file

        Parameters
        ----------
        file_path
            target txt file path
        content
            text string
        """
        with open(file_path, 'a', encoding="utf-8", errors="ignore") as f:
            f.write(content)

    def string_to_hex(self, utf_8_string: str) -> str:
        return utf_8_string.encode("utf-8", errors="ignore").hex()

    def hex_to_string(self, hex_string: str) -> str:
        return bytes.fromhex(hex_string).decode("utf-8", errors="ignore")

    def __make_sure_txt_exist(self, path: str):
        if not os.path.exists(path):
            self.write(path, "")

    def read_settings(self, key: str, defult: str) -> str:
        try:
            settings_path = os.path.join(self.current_dir, 'settings.ini')
            self.__make_sure_txt_exist(settings_path)
            text = self.read(settings_path)
            data = json.loads(text)
            return data[key]
        except Exception as e:
            print(e)
            return defult

    def write_settings(self, key: str, value: str) -> bool:
        try:
            settings_path = os.path.join(self.current_dir, 'settings.ini')
            self.__make_sure_txt_exist(settings_path)
            text = self.read(settings_path)
            try:
                data = json.loads(text)
            except Exception as e:
                print(e)
                data: dict[Any, Any] = dict()
            data.update({key: value})
            text = json.dumps(data)
            self.write(settings_path, text)
            return True
        except Exception as e:
            print(e)
            return False

    def empty_settings(self):
        settings_path = os.path.join(self.current_dir, 'settings.ini')
        try:
            self.write(settings_path, "")
            os.remove(settings_path)
        except Exception as e:
            print(e)

    def log(self, text: str):
        text = str(text)
        now = time.asctime(time.localtime(time.time()))
        text = '\n' * 2 + text + '   ' + '({})'.format(now)
        self.append(self.__log_path, text)

    def get_logs(self):
        return self.read(self.__log_path)


class MyIO():
    def string_to_md5(self, text: str):
        result = hashlib.md5(text.encode())
        return result.hexdigest()

    def base64_to_bytesio(self, base64_string: str):
        img_data = base64.b64decode(base64_string)
        return io.BytesIO(img_data)

    def bytesio_to_base64(self, bytes_io: io.BytesIO):
        bytes_io.seek(0)
        return base64.b64encode(bytes_io.getvalue()).decode()

    def hex_to_bytes(self, hex_string: str):
        return bytes.fromhex(hex_string)

    def bytes_to_hex(self, bytes_data: bytes):
        return bytes_data.hex()




from .app_store_objects import *


from typing import Any
import json
from auto_everything.http_ import Yingshaoxo_Http_Server, Yingshaoxo_Http_Request


class Service_app_store:
    def add_app(self, headers: dict[str, str], item: Add_App_Request) -> Add_App_Response:
        default_response = Add_App_Response()

        try:
            pass
        except Exception as e:
            print(f"Error: {e}")
            #default_response.error = str(e)
            #default_response.success = False

        return default_response

    def search_app(self, headers: dict[str, str], item: Search_App_Request) -> Search_App_Response:
        default_response = Search_App_Response()

        try:
            pass
        except Exception as e:
            print(f"Error: {e}")
            #default_response.error = str(e)
            #default_response.success = False

        return default_response

    def get_app_detail(self, headers: dict[str, str], item: Get_App_Detail_Request) -> Get_App_Detail_Response:
        default_response = Get_App_Detail_Response()

        try:
            pass
        except Exception as e:
            print(f"Error: {e}")
            #default_response.error = str(e)
            #default_response.success = False

        return default_response

    def export_data(self, headers: dict[str, str], item: Export_Data_Request) -> Export_Data_Response:
        default_response = Export_Data_Response()

        try:
            pass
        except Exception as e:
            print(f"Error: {e}")
            #default_response.error = str(e)
            #default_response.success = False

        return default_response


def run(service_instance: Service_app_store, port: str, html_folder_path: str="", serve_html_under_which_url: str="/"):
    def handle_get_url(sub_url: str, headers: dict[str, str]) -> str:
        return 'Hi there, this website is using yrpc (Yingshaoxo remote procedure control module).'

    def handle_post_url(sub_url: str, headers: dict[str, str], item: dict[str, Any]) -> dict | str:
        sub_url = sub_url.strip("/")
        sub_url = sub_url.replace("app_store", "", 1)
        sub_url = sub_url.strip("/")
        request_url = sub_url.split("/")[0].strip()

        if (request_url == ""):
            return f"Request url '{request_url}' is empty"
        elif (request_url == "add_app"):
            correct_item = Add_App_Request().from_dict(item)
            return json.dumps((service_instance.add_app(headers, correct_item)).to_dict())

        elif (request_url == "search_app"):
            correct_item = Search_App_Request().from_dict(item)
            return json.dumps((service_instance.search_app(headers, correct_item)).to_dict())

        elif (request_url == "get_app_detail"):
            correct_item = Get_App_Detail_Request().from_dict(item)
            return json.dumps((service_instance.get_app_detail(headers, correct_item)).to_dict())

        elif (request_url == "export_data"):
            correct_item = Export_Data_Request().from_dict(item)
            return json.dumps((service_instance.export_data(headers, correct_item)).to_dict())

        return f"No API url matchs '{request_url}'"

    def general_handler(request: Yingshaoxo_Http_Request) -> dict | str:
        response = f"No handler for {request.url}"
        if request.method == "GET":
            response = handle_get_url(request.url, request.headers)
        elif request.method == "POST":
            response = handle_post_url(request.url, request.headers, json.loads(request.payload))
        return response

    router = {
        r"(.*)": general_handler,
    }

    yingshaoxo_http_server = Yingshaoxo_Http_Server(router=router)
    yingshaoxo_http_server.start(host="0.0.0.0", port=int(port), html_folder_path=html_folder_path, serve_html_under_which_url=serve_html_under_which_url)


if __name__ == "__main__":
    service_instance = Service_app_store()
    run(service_instance, port="6060")


from __future__ import annotations
import string as built_in_string_module


class String:
    def __init__(self):
        pass

    def capitalize_the_first_character_of_a_string(self, text: str) -> str:
        if len(text) == 0:
            return text
        return text[0].capitalize() + text[1:]
    
    def remove_all_special_characters_from_a_string(self, 
        text: str, 
        allow_numbers: bool = True, 
        allow_ascii_lowercase_characters: bool = True, 
        allow_ascii_uppercase_characters: bool = True, 
        allow_whitespace: bool = False,
        allow_punctuation: bool = False, 
        white_list_characters: str = ''
    ) -> str:
        new_text = ""
        for char in text:
            added = False

            if allow_numbers:
                if char in built_in_string_module.digits:
                    new_text += char
                    added = True
            if allow_ascii_lowercase_characters:
                if char in built_in_string_module.ascii_lowercase:
                    new_text += char
                    added = True
            if allow_ascii_uppercase_characters:
                if char in built_in_string_module.ascii_uppercase:
                    new_text += char
                    added = True
            if allow_whitespace:
                if char in built_in_string_module.whitespace:
                    new_text += char
                    added = True
            if allow_punctuation:
                if char in built_in_string_module.punctuation:
                    new_text += char
                    added = True

            if added == False:
                if char in white_list_characters:
                    new_text += char
                    added = True

        return new_text



#!/usr/bin/env /home/yingshaoxo/anaconda3/bin/python3
import os

from auto_everything.base import Python, Terminal
from auto_everything.develop import YRPC
from auto_everything.disk import Disk
from auto_everything.database import Database_Of_Yingshaoxo


py = Python()
t = Terminal()
yrpc = YRPC()
disk = Disk()


class Tools():
    def __init__(self) -> None:
        self.project_root_folder = disk.get_directory_path(os.path.realpath(os.path.abspath(__file__)))

    def generate_protocol_files(self):
        input_folder=disk.join_paths(self.project_root_folder, "./protocol")
        input_files = ["app_store.proto"]
        output_folder_for_python=disk.join_paths(self.project_root_folder, "./back_end/generated_yrpc")
        output_folder_for_javascript=disk.join_paths(self.project_root_folder, "./front_end/src/generated_yrpc")

        yrpc.generate_code(
            which_language="typescript",
            input_folder=input_folder,
            input_files=input_files,
            output_folder=output_folder_for_javascript
        )

        yrpc.generate_code(
            which_language="python",
            input_folder=input_folder,
            input_files=input_files,
            output_folder=output_folder_for_python
        )

        Database_Of_Yingshaoxo.generate_code_from_yrpc_protocol(
            which_language="python",
            input_folder=input_folder,
            input_files=input_files,
            output_folder=output_folder_for_python
        )

    def build_front_end(self):
        t.run(f"""
        cd {self.project_root_folder}
        cd front_end
        npm i
        npm run build
        #rm -fr ../back_end/vue/*
        #mkdir -p ../back_end/vue/*
        #cp -fr dist/* ../back_end/vue/
        """)

    # def rebuild_docker_image(self):
    #     # self.build_front_end()
    #     t.run(f"""
    #     cd {self.project_root_folder}
    #     docker-compose -f docker-compose.service.yaml down
    #     docker rmi yingshaoxo/it_has_alternatives
    #     docker-compose -f docker-compose.service.yaml up -d
    #     """)


py.make_it_runnable()
py.fire(Tools)



from dataclasses import dataclass
from typing import Any, Callable
import re
import asyncio

from auto_everything.network import Network
net_work = Network()


class MyO365():
    def __init__(self, credentials: tuple[str, str]):
        """
        credentials = ('dhasldl-b8ab-4730-93f1-cee730a4044b', 'salhdlghsaldh~ufOj2-4~69sCMZ05D_')
        """
        from O365 import Account
        self.account = Account(credentials)
        ok = False
        if not self.account.is_authenticated:
            ok = self.account.authenticate(scopes=['basic', 'message_all']) #type: ignore
        else:
            ok = True
        if ok:
            print('Authenticated!')
        else:
            print('Not Authenticated!')
            raise Exception("Email sender not authenticated!")

    def send_email(self, receiver: str, subject: str, body: str):
        m = self.account.new_message() #type: ignore
        m.to.add(receiver) #type: ignore
        m.subject = subject
        m.body = body
        m.send()


class SMTP_Service():
    domain_to_ip_list: dict[str, Any] = {}

    def __init__(self, host: str, port: int, handler: Callable[[str, str, list[str], str], None], auth_ip_source: bool = False):
        """
        host: 0.0.0.0
        port: 25
        handler: (from_ip: str, from: str, to: list[str], content: str) => None
        """
        from aiosmtpd.controller import Controller

        class CustomHandler:
            async def handle_DATA(self, server: Any, session: Any, envelope: Any):
                mail_from = envelope.mail_from
                mail_to = envelope.rcpt_tos
                data = envelope.content
                source_ip_address = session.peer[0]

                if (auth_ip_source == True):
                    ok = SMTP_Service.check_if_an_email_was_sent_from_a_domain(email_address=mail_from, source_ip=source_ip_address)
                    if not ok:
                        return

                try:
                    # print(f"mail_from: {mail_from}")
                    # print("\n\n---\n\n")
                    handler(source_ip_address, mail_from, mail_to, data.decode(encoding="utf-8"))
                except Exception as e:
                    print(e)
                    return '500 Could not process your message'

                return '250 OK'
        
        self.auth_ip_source = auth_ip_source
        self.host = host
        self.port = port
        self.custom_handler = CustomHandler()
        self.controller = Controller(self.custom_handler, hostname=host, port=port)

    def start(self):
        from time import sleep
        self.controller.start()
        print(f'SMTP server is running on {self.host}:{self.port}.\n\n')
        while True:
            sleep(5)
    
    def stop(self):
        self.controller.stop()
    
    @staticmethod
    def get_title_from_email_string_data(data: str) -> None | str:
        title_list = re.findall(r"Subject:\s+(.*)", data, re.IGNORECASE)
        if len(title_list) == 0:
            return None
        else:
            return title_list[0]
    
    @staticmethod
    def get_authorized_ip_list_from_an_email_domain(email_address: str) -> list[str]:
        default_result: list[str] = []

        base_domain = email_address
        if "@" in email_address:
            base_domain = email_address.split("@")[1].strip()
        
        if base_domain in SMTP_Service.domain_to_ip_list.keys():
            return SMTP_Service.domain_to_ip_list[base_domain]
        
        text_record_list = net_work.get_text_record_by_using_domain_url(url=base_domain)
        found = None
        for one_record in text_record_list:
            if "v=spf1" in one_record:
                found = one_record
                break
        
        if found != None:
            ipv4_list = re.findall(r"ip4:([\w\.\-\_\/]+)", found)
            if len(ipv4_list) > 0:
                return ipv4_list

            result: list[str] = []

            rediret_list = re.findall(r"redirect:([\w\.\-\_\/]+)", found)
            if len(rediret_list) > 0:
                result += SMTP_Service.get_authorized_ip_list_from_an_email_domain(email_address=rediret_list[0])

            include_list = re.findall(r"include:([\w\.\-\_\/]+)", found)
            for one in include_list:
                result += SMTP_Service.get_authorized_ip_list_from_an_email_domain(email_address=one)
            
            if len(result) > 0:
                SMTP_Service.domain_to_ip_list[base_domain] = result

            return result

        return default_result

    @staticmethod
    def check_if_an_email_was_sent_from_a_domain(email_address: str, source_ip: str) -> bool:
        network_or_ip_list = SMTP_Service.get_authorized_ip_list_from_an_email_domain(email_address=email_address)
        for one in network_or_ip_list:
            yes = net_work.check_if_an_ip_in_an_ip_network(ip=source_ip, ip_network=one)
            if yes:
                return True
        return False


@dataclass()
class My_Telegram_Message():
    from_chat_id: str | None = None
    from_user_id: str | None = None
    from_user_name: str | None = None
    text: str | None = None


class Telegram_Bot():
    def __init__(self, token: str):
        try:
            import telegram
            self.telegram = telegram
            self.bot: telegram.Bot = telegram.Bot(token)
            self.last_update_id = None
        except Exception as e:
            print(e)
            print(f"""
error: You should install python-telegram-bot by using:

python -m pip install python-telegram-bot
            """)
    
    def send_message(self, chat_id: str, text: str):
        async def send_message_function():
            async with self.bot:
                await self.bot.send_message(text=text, chat_id=int(chat_id)) #type: ignore
        loop = asyncio.get_event_loop()
        task = loop.create_task(send_message_function())
        loop.run_until_complete(task)
        
    async def get_new_message_updates(self):
        messages: list[self.telegram.Update] = []
        if self.last_update_id != None:
            messages = list(await self.bot.get_updates(offset=self.last_update_id+1, timeout=30)) #type: ignore
        else:
            messages = list(await self.bot.get_updates(timeout=30)) #type: ignore
        if len(messages) > 0:
            self.last_update_id = messages[-1].update_id
        return messages

    async def get_messages(self) -> list[My_Telegram_Message]:
        messages: list[My_Telegram_Message] = []
        updates = await self.get_new_message_updates()
        for update in updates:
            if update.message != None and update.message.from_user != None and update.message.from_user.username != None and update.message.text != None:
                """
                Message(channel_chat_created=False, chat=Chat(first_name='yingshao', id=131513300, last_name='xo', type=<ChatType.PRIVATE>, username='yingshaoxo'), date=datetime.datetime(2023, 3, 30, 9, 4, 23, tzinfo=<UTC>), delete_chat_photo=False, from_user=User(first_name='yingshao', id=131513300, is_bot=False, language_code='en', last_name='xo', username='yingshaoxo'), group_chat_created=False, message_id=1363, supergroup_chat_created=False, text='hi')
                """
                messages.append(My_Telegram_Message(
                    from_chat_id=str(update.message.chat.id),
                    from_user_id=str(update.message.from_user.id),
                    from_user_name=str(update.message.from_user.username),
                    text=str(update.message.text)
                ))
        return messages

    def get_message_loop(self, new_message_handler: Callable[[My_Telegram_Message], None], sleep_time_in_second: float = 1):
        """
        new_message_handler: (My_Telegram_Message) => None
            it is a function that handles new message, it will receive an object like this:
            ```
            @dataclass()
            class My_Telegram_Message():
                from_chat_id: str | None = None
                from_user_id: str | None = None
                from_user_name: str | None = None
                text: str | None = None
            ```
        """
        async def loop_function():
            while True:
                messages = await self.get_messages()
                for msg in messages:
                    try:
                        # print(msg)
                        new_message_handler(msg)
                    except Exception as e:
                        print(f"error: {e}")
                await asyncio.sleep(sleep_time_in_second)
        loop = asyncio.get_event_loop()
        task = loop.create_task(loop_function())
        loop.run_until_complete(task)
        # loop = asyncio.new_event_loop()
        # asyncio.set_event_loop(loop)
        # asyncio.ensure_future(loop_function())
        # loop.run_forever()


if __name__ == '__main__':
    # def handle_email(from_id: str, from_: str, to: list[str], message: str):
    #     print(from_)
    #     print(to)
    #     print(message)

    # smtp_service = SMTP_Service(
    #     host="0.0.0.0",
    #     port=25,
    #     handler=handle_email
    # )

    # # smtp_service.start()

    # from_email = "ddsd@protonmail.com"
    # result = SMTP_Service.get_authorized_ip_list_from_an_email_domain(from_email)
    # if (len(result) > 0):
    #     an_ip = result[0].split("/")[0]
    #     ok = SMTP_Service.check_if_an_email_was_sent_from_a_domain(email_address=from_email, source_ip=an_ip)
    #     print(ok)
    #     ok = SMTP_Service.check_if_an_email_was_sent_from_a_domain(email_address=from_email, source_ip="127.0.0.1")
    #     print(ok)
    #     ok = SMTP_Service.check_if_an_email_was_sent_from_a_domain(email_address=from_email, source_ip=an_ip)
    #     print(ok)

    # from auto_everything.my_email import Telegram_Bot
    # telegram_bot = Telegram_Bot(token="")

    # def message_handler(msg_object):
    #     print(msg_object.text)

    # if __name__ == '__main__':
    #     telegram_bot.get_message_loop(
    #         new_message_handler=message_handler
    #     )

    pass


from typing import Any, Callable
from dataclasses import dataclass

import os
import socket
import json
import re
from time import sleep


@dataclass()
class Yingshaoxo_Http_Request():
    context: Any
    host: str
    method: str
    url: str
    url_arguments: dict[str, str]
    headers: dict[str, str]
    payload: str | None


def _handle_socket_request(socket_connection, context, router, handle_get_file_url):
    try:
        host = None
        method = None
        url = None
        url_arguments = dict()
        http_standards = None

        raw_http_request_bytes = socket_connection.recv(1024) # one utf-8 char is 0~4 bytes, that's why for the following code, I times the length by 4 to make sure we receive all data
        raw_http_request = raw_http_request_bytes.decode("utf-8", errors="ignore")
        #print(raw_http_request)
        #print(repr(raw_http_request))

        splits = raw_http_request.strip().split("\n")
        if (len(splits) > 0):
            head_line = splits[0]
            head_line_splits = head_line.split(" ")
            if len(head_line_splits) == 3:
                method, url, http_standards = head_line_splits
                url_splits = url.split("?")
                if len(url_splits) >= 2:
                    url = url_splits[0]
                    raw_url_arguments = "?".join(url_splits[1:])
                    url_arguments_splits = raw_url_arguments.split("&")
                    for one in url_arguments_splits:
                        if "=" in one:
                            argument_key, argument_value = one.split("=")
                            url_arguments[argument_key] = argument_value
                else:
                    pass

        if (method == None or url == None or http_standards == None):
            print(f"Unkonw http request:\n{raw_http_request}")
            exit()
        else:
            pass

        if (method == "OPTIONS"):
            response = f"""
HTTP/1.1 200 OK
Access-Control-Allow-Origin: *
Access-Control-Allow-Methods: *
Access-Control-Allow-Headers: *\r\n\r\ndone
""".strip()
            response = response.encode("utf-8", errors="ignore")
            socket_connection.sendall(response)
            socket_connection.shutdown(1)
            socket_connection.close()
            exit()

        raw_headers_lines = splits[1:]
        raw_headers_lines = [line for line in raw_headers_lines if ": " in line]
        headers_dict = {}
        for line in raw_headers_lines:
            header_splits = line.split(": ")
            key = header_splits[0]
            value = ":".join(header_splits[1:])
            headers_dict[key] = value

        content_length = None
        payload = None
        payload_seperator_bytes = b"\r\n\r\n"
        payload_seperator = "\r\n\r\n"
        if "Content-Length" in headers_dict:
            content_length = int(headers_dict["Content-Length"])

        if content_length != None:
            # receive the rest of data by using socket receiv
            payload_splits = raw_http_request_bytes.split(payload_seperator_bytes)
            if len(payload_splits) >= 2:
                # already has all headers, need payload
                payload = payload_splits[1]
                if len(payload) >= content_length:
                    pass
                else:
                    payload += socket_connection.recv((content_length)*4-len(payload)+200)
                    payload = payload.decode("utf-8", errors="ignore")
            else:
                # missing some headers, need more data, including payload
                raw_http_request_bytes += socket_connection.recv((content_length)*4+len(raw_http_request_bytes)+200)
                raw_http_request = raw_http_request_bytes.decode("utf-8", errors="ignore")
                payload = raw_http_request_bytes.split(payload_seperator_bytes)[1]
                payload = payload.decode("utf-8", errors="ignore")

        # do the process directly
        splits = raw_http_request.split(payload_seperator)
        header_line_list = splits[0].split("\n")[1:]
        headers_dict = {}
        for line in header_line_list:
            header_splits = line.split(": ")
            key = header_splits[0]
            value = (":".join(header_splits[1:])).strip()
            headers_dict[key] = value
        if "Host" in headers_dict:
            host = headers_dict["Host"]

        print(host, method, url)
        #print(f"headers:\n{headers_dict}")
        #print(f"payload:\n{payload}")
        raw_response = None
        response = f"HTTP/1.1 500 Server error\r\n\r\n".lstrip()

        if handle_get_file_url != None and method == "GET":
            # handle file download request, for example, html, css...
            raw_response = handle_get_file_url(url)

        # if do not need to serve file, or file not exists, then handle others
        if raw_response == None:
            the_request_object = Yingshaoxo_Http_Request(
                context=context,
                host=host,
                method=method,
                url=url,
                url_arguments=url_arguments,
                headers=headers_dict,
                payload=payload
            )
            for route_regex_expression, route_function in reversed(list(router.items())):
                if re.fullmatch(route_regex_expression, url) != None:
                    raw_response = route_function(the_request_object)

        if type(raw_response) == str:
            response = f"""
HTTP/1.1 200 OK
Access-Control-Allow-Origin: *\r\n\r\n{raw_response}
""".strip()
        elif type(raw_response) == dict:
            raw_response = json.dumps(raw_response, indent=4)
            json_length = len(raw_response)
            response = f"""
HTTP/1.1 200 OK
Content-Type: application/json
Content-Length: {json_length}
Access-Control-Allow-Origin: *\r\n\r\n{raw_response}
            """.strip()
        elif type(raw_response) == bytes:
            bytes_length = len(raw_response)

            the_content_type = None
            if url.endswith(".html"):
                the_content_type = "text/html"
            elif url.endswith(".css"):
                the_content_type = "text/css"
            elif url.endswith(".js"):
                the_content_type = "text/javascript"

            if the_content_type != None:
                response = f"""
HTTP/1.1 200 OK
Content-Type: {the_content_type}
Content-Length: {bytes_length}
Access-Control-Allow-Origin: *\r\n\r\n""".lstrip()
            else:
                response = f"""
HTTP/1.1 200 OK
Content-Length: {bytes_length}
Access-Control-Allow-Origin: *\r\n\r\n""".lstrip()
            response = response.encode("utf-8", errors="ignore")
            response += raw_response
        else:
            response = f"HTTP/1.1 500 Server error\r\n\r\nNo router for {url}".strip()

        if type(response) == str:
            response = response.encode("utf-8", errors="ignore")

        socket_connection.sendall(response)
    except Exception as e:
        print(e)
        response = f"HTTP/1.1 200 OK\r\n\r\n{e}".strip()
        response = response.encode("utf-8", errors="ignore")
        socket_connection.sendall(response)
    finally:
        socket_connection.shutdown(1)
        socket_connection.close()
        exit()


def _yingshaoxo_home_handler_example(request: Yingshaoxo_Http_Request) -> dict:
    return {"message": "Hello, world, fight for inner peace."}

def _yingshaoxo_special_handler_example(request: Yingshaoxo_Http_Request) -> dict:
    return "Hello, world, fight for personal freedom."

_yingshaoxo_router_example = {
    r"/freedom": _yingshaoxo_special_handler_example,
    r"(.*?)": _yingshaoxo_home_handler_example,
}


class Yingshaoxo_Http_Server():
    def __init__(self, router: dict[str, Callable[[Yingshaoxo_Http_Request], str|dict]]):
        """
        router: a dict where key is the url regex, value is a function like "def handle_function(request: Yingshaoxo_Http_Request) -> str|dict"
        """
        import multiprocessing
        self._multiprocessing = multiprocessing
        multiprocess_manager = self._multiprocessing.Manager()

        self.context = multiprocess_manager.dict()
        self.router = router

    def start(self, host:str = "0.0.0.0", port:int = 80, html_folder_path: str="", serve_html_under_which_url: str="/"): 
        try:
            handle_get_file_url = None
            if (html_folder_path != ""):
                if os.path.exists(html_folder_path) and os.path.isdir(html_folder_path):
                    def handle_get_file_url(sub_url: str) -> bytes | None:
                        sub_url = sub_url.strip("/")
                        sub_url = sub_url.lstrip(serve_html_under_which_url)
                        if sub_url == '':
                            sub_url = 'index.html'
                        real_file_path = f"{os.path.join(html_folder_path, sub_url)}"
                        if os.path.exists(real_file_path) and os.path.isfile(real_file_path):
                            with open(real_file_path, mode="rb") as f:
                                the_data = f.read()
                        else:
                            #return b"Resource not found"
                            return None
                        return the_data
                else:
                    print(f"Error: You should give me an absolute html_folder_path than {html_folder_path}")

            server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server.bind((host, port))
            server.listen(1)

            print(f"Service is on http://{host}:{port}")

            process_list = []

            while True:
                socket_connection, addr = server.accept()
                process = self._multiprocessing.Process(target=_handle_socket_request, args=(socket_connection, self.context, self.router, handle_get_file_url))
                process.start()
                process_list.append(process)

                new_process_list = []
                for a_process in process_list:
                    if a_process.is_alive():
                        new_process_list.append(a_process)
                process_list = new_process_list
        except KeyboardInterrupt:
            print("Quit...")
            for a_process in process_list:
                if a_process.is_alive():
                    a_process.terminate()
            server.shutdown(1)
            server.close()
        except Exception as e:
            print(e)
            print("Quit...")
            for a_process in process_list:
                if a_process.is_alive():
                    a_process.terminate()
            server.shutdown(1)
            server.close()
        finally:
            pass


class Yingshaoxo_Threading_Based_Http_Server():
    def __init__(self, router: dict[str, Callable[[Yingshaoxo_Http_Request], str|dict]]):
        """
        router: a dict where key is the url regex, value is a function like "def handle_function(request: Yingshaoxo_Http_Request) -> str|dict"
        """
        from http.server import HTTPServer, BaseHTTPRequestHandler
        from socketserver import ThreadingMixIn
        #import http.server as built_in_http_server
        self._HTTPServer = HTTPServer
        self._BaseHTTPRequestHandler = BaseHTTPRequestHandler
        self._ThreadingMixIn = ThreadingMixIn

        self.context = dict()
        self.router = router

    def _get_headers_dict_from_string(self, headers: str) -> dict:
        dic = {}
        for line in headers.split("\n"):
            if line.startswith(("GET", "POST")):
                continue
            point_index = line.find(":")
            dic[line[:point_index].strip()] = line[point_index+1:].strip()
        return dic

    def start(self, host:str = "0.0.0.0", port: int = 80, html_folder_path: str="", serve_html_under_which_url: str="/"):
        def handle_file_request_url(sub_url: str) -> bytes | None:
            return b'Hi there, this website is using yrpc (Yingshaoxo remote procedure control module).'

        if (html_folder_path != ""):
            if os.path.exists(html_folder_path) and os.path.isdir(html_folder_path):
                def handle_file_request_url(sub_url: str) -> bytes | None:
                    sub_url = sub_url.strip("/")
                    sub_url = sub_url.lstrip(serve_html_under_which_url)
                    if sub_url == '':
                        sub_url = 'index.html'
                    real_file_path = f"{os.path.join(html_folder_path, sub_url)}"
                    if os.path.exists(real_file_path) and os.path.isfile(real_file_path):
                        with open(real_file_path, mode="rb") as f:
                            the_data = f.read()
                    else:
                        #return None #instead return None, return index.html for single app page
                        sub_url = 'index.html'
                        real_file_path = f"{os.path.join(html_folder_path, sub_url)}"
                        with open(real_file_path, mode="rb") as f:
                            the_data = f.read()
                    return the_data
            else:
                print(f"Error: You should give me an absolute html_folder_path than {html_folder_path}")

        def handle_any_url(method: str, sub_url: str, headers: dict[str, str], payload: dict[str, Any] | None = None) -> tuple[bytes, str | bytes | dict]:
            #sub_url = sub_url.strip("/")
            #sub_url = sub_url.replace("{identity_name}", "", 1)
            #sub_url = sub_url.strip("/")
            #request_url = sub_url.split("/")[0].strip()

            raw_response = None

            if method == "GET":
                raw_response = handle_file_request_url(sub_url)

            url_arguments = dict()
            url_splits = sub_url.split("?")
            if len(url_splits) >= 2:
                sub_url = url_splits[0]
                raw_url_arguments = "?".join(url_splits[1:])
                url_arguments_splits = raw_url_arguments.split("&")
                for one in url_arguments_splits:
                    if "=" in one:
                        argument_key, argument_value = one.split("=")
                        url_arguments[argument_key] = argument_value

            the_request_object = Yingshaoxo_Http_Request(
                context=self.context,
                host=headers.get("Host"),
                method=method,
                url=sub_url,
                url_arguments=url_arguments,
                headers=headers,
                payload=payload
            )
            for route_regex_expression, route_function in reversed(list(self.router.items())):
                if re.fullmatch(route_regex_expression, sub_url) != None:
                    raw_response = route_function(the_request_object)

            if raw_response == None:
                raw_response = f"No API url matchs '{sub_url}'"

            raw_type = str
            if type(raw_response) == str:
                raw_response = raw_response.encode("utf-8", errors="ignore")
                raw_type = str
            elif type(raw_response) == dict:
                raw_response = json.dumps(raw_response, indent=4).encode("utf-8", errors="ignore")
                raw_type = dict
            elif type(raw_response) == bytes:
                raw_type = bytes

            return raw_response, raw_type

        class WebRequestHandler(self._BaseHTTPRequestHandler):
            def do_OPTIONS(self2):
                self2.send_response(200, "ok")
                self2.send_header('Access-Control-Allow-Origin', '*')
                self2.send_header('Access-Control-Allow-Methods', '*')
                self2.send_header("Access-Control-Allow-Headers", "*")
                self2.end_headers()

            def do_GET(self2):
                sub_url = self2.path
                headers = self._get_headers_dict_from_string(self2.headers.as_string())

                self2.send_response(200)

                self2.send_header("Access-Control-Allow-Origin", "*")

                if sub_url.endswith(".html"):
                    self2.send_header("Content-Type", "text/html")
                elif sub_url.endswith(".css"):
                    self2.send_header("Content-Type", "text/css")
                elif sub_url.endswith(".js"):
                    self2.send_header("Content-Type", "text/javascript")

                response, raw_type = handle_any_url("GET", sub_url, headers, None)

                self2.end_headers()
                self2.wfile.write(response)

            def do_POST(self2):
                sub_url = self2.path
                headers = self._get_headers_dict_from_string(self2.headers.as_string())

                content_length = headers.get('Content-Length')
                if content_length is None:
                    self2.wfile.write("What you send is not json".encode("utf-8", errors="ignore"))
                    return
                else:
                    content_length = int(content_length)

                if content_length == 0:
                    self2.wfile.write("What you send is not json".encode("utf-8", errors="ignore"))
                    return

                request_json_dict = json.loads(self2.rfile.read(content_length))

                self2.send_response(200)
                self2.send_header("Access-Control-Allow-Origin", "*")

                response, raw_type = handle_any_url("POST", sub_url, headers, request_json_dict)
                if raw_type == dict:
                    self2.send_header("Content-Type", "application/json")

                self2.end_headers()
                self2.wfile.write(response)

        class ThreadedHTTPServer(self._ThreadingMixIn, self._HTTPServer):
            pass

        # Setting TCP Address
        server_address = (host, port)

        # invoking server
        http = ThreadedHTTPServer(server_address, WebRequestHandler)

        print(f"The website is running at: http://127.0.0.1:{port}/")

        http.serve_forever()


def run_a_command_with_hot_load(watch_path: str, hotload_command: str):
    """
    watch_path: a folder you want to watch, whenever some of those file get changed, the hotload_command will get re executed
    hotload_command: a bash command to start the server, for example, "python3 main.py"

    You should run this function on a diffirent python file than the hotload_command has.
    """
    import multiprocessing
    from auto_everything.develop import Develop
    from auto_everything.terminal import Terminal
    develop = Develop()
    terminal = Terminal()

    def run_the_process():
        terminal.run(hotload_command, wait=True)

    the_running_process = multiprocessing.Process(target=run_the_process, args=())
    the_running_process.start()

    while True:
        changed = develop.whether_a_folder_has_changed(folder_path=watch_path, type_limiter=[".py", ".html", ".css", ".js"])
        if (changed):
            print("Source code get changed, doing a reloading now...")
            the_running_process.kill()
            while the_running_process.is_alive():
                sleep(1)
            the_running_process = multiprocessing.Process(target=run_the_process, args=())
            the_running_process.start()
        sleep(1)


if __name__ == "__main__":
    #yingshaoxo_http_server = Yingshaoxo_Http_Server(router=_yingshaoxo_router_example)
    yingshaoxo_http_server = Yingshaoxo_Threading_Based_Http_Server(router=_yingshaoxo_router_example)
    yingshaoxo_http_server.start(port=1212, html_folder_path="./")



from auto_everything.io import IO
from auto_everything.terminal import Terminal, Terminal_User_Interface
from auto_everything.disk import Disk
from auto_everything.python import Python
from auto_everything.network import Network
from auto_everything.cryptography import Encryption_And_Decryption
from auto_everything.develop import YRPC

__all__ = [
    'IO',
    'Terminal',
    "Terminal_User_Interface",
    'Disk',
    'Python',
    'Network',
    'Encryption_And_Decryption',
    'YRPC'
]


#!/usr/bin/env /usr/bin/python3
from typing import Any

import os
import sys
import platform
import json
from time import sleep

try:
    if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
        # in pyinstaller
        def resource_path(relative_path: str) -> str:
            if hasattr(sys, '_MEIPASS'):
                return os.path.join(sys._MEIPASS, relative_path) #type: ignore
            return os.path.join(os.path.abspath("."), relative_path)
        _the_current_path = resource_path(".")
        sys.path.append(os.path.abspath(_the_current_path.replace("/./", "/")))
    else:
        # normal process
        sys.path.append(os.path.dirname(os.path.abspath(__file__)))
except Exception as e:
    pass

from auto_everything.io import IO
from auto_everything.disk import Disk
from auto_everything.python import Python
from auto_everything.terminal import Terminal, Terminal_User_Interface
try:
    from auto_everything.string_ import String
except Exception as e:
    from auto_everything.string import String


py = Python()
io_ = IO()
terminal = Terminal()
disk = Disk()
python_ = Python()
terminal_user_interface = Terminal_User_Interface()
string_tool = String()


class Tools():
    def __init__(self) -> None:
        # get static resource folder
        if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
            #print('running in a PyInstaller bundle')
            def resource_path(relative_path: str) -> str:
                if hasattr(sys, '_MEIPASS'):
                    return os.path.join(sys._MEIPASS, relative_path) #type: ignore
                return os.path.join(os.path.abspath("."), relative_path)
            resource_basic_folder_path = resource_path(".")
        else:
            #print('running in a normal Python process')
            resource_basic_folder_path = disk.get_directory_path(__file__)
        self.resource_basic_folder_path = disk.join_paths(resource_basic_folder_path, 'resources').replace("/./", "/")
        self.auto_everything_basic_folder_path = disk.join_paths(resource_basic_folder_path, 'auto_everything').replace("/./", "/")

        # get python executable path
        self.host_python_executable_path = "python"
        if " 3." in terminal.run_command("python --version"):
            self.host_python_executable_path = "python"
        elif " 3." in terminal.run_command("python3 --version"):
            self.host_python_executable_path = "python3"

        # get git user email and username
        self.git_user_email = terminal.run_command("git config user.email").strip().split('\n')[0]
        self.git_user_name = terminal.run_command("git config user.name").strip().split('\n')[0].lower()
        if self.git_user_name == "":
            try:
                self.git_user_name = string_tool.remove_all_special_characters_from_a_string(os.getlogin(), white_list_characters='_').strip().lower()
            except Exception as e:
                print(e)
                self.git_user_name = ""
            if self.git_user_name == "":
                self.git_user_name = platform.system().lower().strip()

        # get project base folder
        self.project_root_folder = os.getcwd()

        # get virtual env folder
        self.virtual_env_folder = disk.join_paths(self.project_root_folder, ".venv")
        # self.env_activate_file_path = disk.join_paths(self.virtual_env_folder, "bin/activate")

        # get package json file
        self.package_json_file_path = disk.join_paths(self.project_root_folder, "package.json")

        # get local dependencies folder
        self.python_local_modules_folder = disk.join_paths(self.project_root_folder, ".python_modules")

    def _add_to_gitignore(self, name: str):
        gitignore_path = disk.join_paths(self.project_root_folder, ".gitignore")
        if disk.exists(gitignore_path):
            text = io_.read(gitignore_path)
            lines = text.split("\n")
            if not any([line.strip() == name for line in lines]):
                text = f"{name}\n" + text
                io_.write(gitignore_path, text)
        else:
            text = name
            io_.write(gitignore_path, text)

    def _add_to_dockerignore(self, name: str):
        gitignore_path = disk.join_paths(self.project_root_folder, ".dockerignore")
        if disk.exists(gitignore_path):
            text = io_.read(gitignore_path)
            lines = text.split("\n")
            if not any([line.strip() == name for line in lines]):
                text = f"{name}\n" + text
                io_.write(gitignore_path, text)
        else:
            text = name
            io_.write(gitignore_path, text)

    def _create_virtual_env(self):
        if disk.exists(self.virtual_env_folder):
            return

        print(f"Creating virtual envirnoment by using '{self.host_python_executable_path}'...")

        terminal.run(f"""
        {self.host_python_executable_path} -m venv {self.virtual_env_folder}
                     """)

        # change permission of activate file
        # terminal.run(f"""chmod 777 {self.env_activate_file_path}""")

        # ignore .venv folder
        self._add_to_gitignore(".venv/")
        self._add_to_gitignore("__pycache__/")
        self._add_to_gitignore(".vscode/")
        self._add_to_gitignore(".github/")
        self._add_to_gitignore("*.swp")

        self._add_to_dockerignore(".venv/")
        self._add_to_dockerignore("__pycache__/")
        self._add_to_dockerignore("*.swp")

        # create .python_modules directory
        if not disk.exists(self.python_local_modules_folder):
            disk.create_a_folder(self.python_local_modules_folder)

    def _get_virtual_env_python_excutable_path(self):
        self._create_virtual_env()

        return disk.join_paths(self.virtual_env_folder, "bin", "python3")

    def _get_virtual_env_pip_path(self):
        self._create_virtual_env()

        return disk.join_paths(self.virtual_env_folder, "bin", "pip3")

    def _get_virtual_env_program(self, name: str, exact_match: bool = False):
        root_search_folder_list = [
            disk.join_paths(self.virtual_env_folder, "bin"),
            disk.join_paths(self.virtual_env_folder, "Scripts"),
        ]
        for search_folder in root_search_folder_list:
            if not disk.exists(search_folder):
                continue
            files = disk.get_files(search_folder, recursive=True)
            for file in files:
                file_name_without_content_after_dot, suffix = disk.get_stem_and_suffix_of_a_file(file)
                if exact_match == False:
                    if name.lower() in file_name_without_content_after_dot.lower():
                        return disk.get_absolute_path(file)
                else:
                    if name == file_name_without_content_after_dot:
                        return disk.get_absolute_path(file)
        return None

    def _hack_into_virtual_env_bash_command(self):
        self._create_virtual_env()

        the_bin_path = disk.join_paths(self.virtual_env_folder, "bin")

        return f"""
        export PATH="{the_bin_path}:$PATH"
        """.strip()

    def _get_package_json_object(self) -> Any:
        try:
            text = io_.read(self.package_json_file_path)
            lines = []
            for line in text.split("\n"):
                if line.strip().startswith("#") or line.strip().startswith("//"):
                    #ignore comments
                    pass
                else:
                    lines.append(line)
            text = "\n".join(lines)
            json_object = json.loads(text)
            return json_object
        except Exception as e:
            print(e)
            return None

    def _get_local_packages_list(self) -> list[str]:
        try:
            return [disk.join_paths("./.python_modules/", path_) for path_ in os.listdir(self.python_local_modules_folder)]
        except Exception as e:
            print(e)
            return []

    def create_a_global_entry_for_this_project(self):
        if not os.environ.get("SUDO_UID") and os.geteuid() != 0:
            print("Sudo permission is needed for this operation.\n")
            print(f"If you do not have root permission but still want to run this project, do this:\ncd {self.project_root_folder} && yppm run")
            return

        binary_version_of_yppm = "/usr/bin/yppm"
        if not disk.exists(binary_version_of_yppm):
            binary_version_of_yppm = f"{self.host_python_executable_path} -m yppm"

        entry_point_bash_code = f"""
#!/bin/sh
cd {self.project_root_folder} && {binary_version_of_yppm} run
        """

        json_object = self._get_package_json_object()
        if "name" not in json_object:
            print("There should have a name in package.json")
            return

        name = json_object.get("name").strip().replace(" ", "_").replace("-", "_")
        io_.write(f"/usr/bin/{name}", entry_point_bash_code)
        terminal.run(f"chmod 777 /usr/bin/{name}")

        print(f"Now you could run this project by using `{name}`\n")
        print(f"If you want to delete this entry, just try `sudo rm /usr/bin/{name}`")

    def create_a_new_project(self):
        global default_template_name, default_project_name

        # select teplates
        default_template_name = "basic_python_project"
        default_template_name = terminal_user_interface.selection_box(
            "Which project template you want to use? ", [
                ("basic_python_project", None),
                ("simple_backend_and_frontend_project", None)
                # ("backend_and_frontend_project", None)
            ]
        )

        # handle project name
        default_project_name = ""

        def assign_name(text: str):
            global default_project_name
            text = text.strip()
            if text == "":
                print("You can't give me an empty name!")
                exit()
            default_project_name = text

        terminal_user_interface.input_box(
            f"\n\nPlease give me the new project name: ",
            default_value=default_project_name,
            handle_function=assign_name
        )

        default_project_name = default_project_name.strip().replace(" ", "_").replace("-", "_")

        # copy template folder
        project_path = disk.join_paths(disk.get_current_working_directory(), default_project_name)
        disk.delete_a_folder(project_path)
        if default_template_name == "basic_python_project":
            source_folder_path = disk.join_paths(self.resource_basic_folder_path, default_template_name)
            disk.copy_a_folder(source_folder_path=source_folder_path, target_folder_path=project_path)

            disk.copy_a_folder(source_folder_path=self.auto_everything_basic_folder_path, target_folder_path=disk.join_paths(project_path, "auto_everything"))

            os.chdir(project_path)
            self.__init__()

            self.init(name=default_project_name)

            print(f"\n\nNow you could go to the new project by using: \ncd {default_project_name}")
        elif default_template_name == "simple_backend_and_frontend_project":
            source_folder_path = disk.join_paths(self.resource_basic_folder_path, default_template_name)
            disk.copy_a_folder(source_folder_path=source_folder_path, target_folder_path=project_path)

            disk.copy_a_folder(source_folder_path=self.auto_everything_basic_folder_path, target_folder_path=disk.join_paths(project_path, "back_end", "auto_everything"))

            os.chdir(project_path)
            self.__init__()

            self.init(name=default_project_name)

            print(f"\n\nNow you could go to the new project by using: \ncd {default_project_name}")
        elif default_template_name == "backend_and_frontend_project":
            pass

    def init(self, name: str = ""):
        # create package.json file in current folder if there does not have one
        if disk.exists(self.package_json_file_path):
            print("There already has a 'package.json' file.")
        else:
            json_content = """
            {
                "name": "",
                "version": "",
                "author": "",
                "main": "main.py",
                "scripts": {},
                "dependencies": [
                ]
            }
            """
            package_object = json.loads(json_content)
            print("\n")

            # handle project name
            name = name.strip()
            if name == "":
                default_project_name = disk.get_directory_name(self.project_root_folder)
                name = terminal_user_interface.input_box(
                    f"Please give me a project name (default '{default_project_name}'): ",
                    default_value=default_project_name,
                    handle_function=None
                )

            package_object["name"] = name.strip().replace(" ", "_").replace("-", "_")

            # handle project version
            default_project_version = "0.0.0"
            def assign_version(text: str):
                package_object["version"] = text.strip()
            terminal_user_interface.input_box(
                f"Please give me a project version (default '{default_project_version}'): ",
                default_value=default_project_version,
                handle_function=assign_version
            )

            # handle author name
            default_author = self.git_user_email
            if "@" not in default_author:
                default_author = ""
            def assign_author(text: str):
                package_object["author"] = text.strip()
            terminal_user_interface.input_box(
                f"Please give me your name (default '{default_author}'): ",
                default_value=default_author,
                handle_function=assign_author
            )

            io_.write(self.package_json_file_path, json.dumps(package_object, indent=4))
            self._create_virtual_env()

            self.install()

            main_py_file_path = disk.join_paths(self.project_root_folder, "main.py")
            if not disk.exists(main_py_file_path):
                io_.write(main_py_file_path, 'print("Hello World!")')

    def run(self, script_name: str = ""):
        if not disk.exists(self.package_json_file_path):
            print("You have to run `yppm init` first.")
            return
        self._create_virtual_env()

        package_object = self._get_package_json_object()

        scripts = package_object.get("scripts")
        if scripts == None:
            print('package.json should have a key called {"scripts": {}}')
            return

        entry_point_python_script = package_object.get("main")
        if entry_point_python_script == None:
            print('package.json should have a key called {"main": "main.py"}')
            return

        script_name = script_name.strip()
        if script_name == "":
            # {self.env_activate_file_path}
            terminal.run(f"""
            {self._hack_into_virtual_env_bash_command()}

            {self._get_virtual_env_python_excutable_path()} {entry_point_python_script}
                         """, use_os_system=True)
        elif script_name == "?":
            if len(scripts.keys()) == 0:
                print("There has no scripts to run.")
            else:
                selections = []
                for script_name in scripts.keys():
                    selections.append(
                        (f'{script_name}', lambda: script_name.strip())
                    )
                selection = terminal_user_interface.selection_box(
                    "Which script you want to run?",
                    selections
                )
                if selection in scripts.keys():
                    terminal.run(f"{scripts[selection]}", use_os_system=True)
        else:
            if script_name in scripts.keys():
                # {self.env_activate_file_path}
                terminal.run(f"""
                {self._hack_into_virtual_env_bash_command()}

                {scripts[script_name]}
                            """, use_os_system=True)
            else:
                print(f"Sorry, script '{script_name}' not exists in the package.json")

    def _install_package(self, package_name: str, upgrade: bool = False):
        pip_path = self._get_virtual_env_pip_path()

        yes_exists = terminal.run_command("yes --version")
        if ("copyright" in yes_exists.strip().lower()):
            yes_exists = True
        else:
            yes_exists = False

        if upgrade == False:
            # {self.env_activate_file_path}
            terminal.run(f"""
            {self._hack_into_virtual_env_bash_command()}
            yes | {pip_path} install {package_name}
                        """)
        else:
            # {self.env_activate_file_path}
            terminal.run(f"""
            {self._hack_into_virtual_env_bash_command()}
            yes | {pip_path} install {package_name} --upgrade
                        """)

    def _uninstall_package(self, package_name: str):
        if (package_name.startswith("./.python_modules")):
            target_folder = disk.join_paths(self.project_root_folder, package_name)
            disk.delete_a_folder(target_folder)
        else:
            pip_path = self._get_virtual_env_pip_path()

            # {self.env_activate_file_path}
            terminal.run(f"""
            {self._hack_into_virtual_env_bash_command()}
            {pip_path} uninstall {package_name} -y
                            """)

    def install(self, package_name: str = ""):
        if not disk.exists(self.package_json_file_path):
            print("You have to run `yppm init` first.")
            return
        self._create_virtual_env()

        package_object = self._get_package_json_object()

        dependencies = package_object.get("dependencies")
        if dependencies == None:
            print('package.json should have a key called {"dependencies": []}')
            return

        package_name = package_name.strip()
        if package_name == "":
            # install all package
            for package_name in dependencies:
                self._install_package(package_name=package_name)
        # elif package_name.startswith("http://") or package_name.startswith("https://") or disk.exists(package_name):
        #     # if without pip, we could use `yppm install https://*.tar.gz` and `yppm install *.tar.gz`
        #     # What yppm will do is download, uncompress, put it into .python_modules/
        #     # Add "./.python_modules/*" into package.json file
        #     pass
        else:
            # install one package
            if package_name not in dependencies:
                # install once
                self._install_package(package_name=package_name)
            else:
                # install again
                self._install_package(package_name=package_name, upgrade=True)

            if package_name not in dependencies:
                package_object["dependencies"].append(package_name)
                io_.write(self.package_json_file_path, json.dumps(package_object, indent=4))

    def uninstall(self, package_name: str = ""):
        if not disk.exists(self.package_json_file_path):
            print("You have to run `yppm init` first.")
            return
        self._create_virtual_env()

        package_object = self._get_package_json_object()

        dependencies = package_object.get("dependencies")
        if dependencies == None:
            print('package.json should have a key called {"dependencies": []}')
            return

        package_name = package_name.strip()
        if package_name == "":
            pass
        elif package_name == "?":
            pip_path = self._get_virtual_env_pip_path()
            package_list_from_pip = terminal.run_command(f"""
                {self._hack_into_virtual_env_bash_command()}
                {pip_path} list
                            """).strip()
            package_list_from_pip = package_list_from_pip.split("\n\n")[0]
            package_list_from_pip = package_list_from_pip.strip().split("\n")[2:]
            package_list_from_pip = [package_name for package_name in package_list_from_pip if package_name.strip()[-1].isdigit()]
            package_list_from_pip = [package_name.split(" ")[0].replace("-", "_") for package_name in package_list_from_pip]

            # add json defined packages on the top of the list
            for package_in_json_file in dependencies:
                if package_in_json_file not in package_list_from_pip:
                    package_list_from_pip.insert(0, package_in_json_file)

            # add local packages on the end of the list
            for package_in_local in self._get_local_packages_list():
                if package_in_local not in package_list_from_pip:
                    package_list_from_pip.append(package_in_local)

            package_name = terminal_user_interface.selection_box("Please select the package you want to uninstall: ", [
                (package_name, None) for package_name in package_list_from_pip
            ])

            self._uninstall_package(package_name=package_name)
        else:
            self._uninstall_package(package_name=package_name)

        if package_name in dependencies:
            package_object["dependencies"] = [one for one in package_object["dependencies"] if one != package_name]
            io_.write(self.package_json_file_path, json.dumps(package_object, indent=4))

    def build(self, arguments: str = ""):
        self._create_virtual_env()
        python_path = self._get_virtual_env_python_excutable_path()
        pip_path = self._get_virtual_env_pip_path()

        package_object = self._get_package_json_object()

        name = package_object.get("name")
        if name == None:
            print('package.json should have a key called {"name": "your_app_name"}')
            return

        entry_point_python_script = package_object.get("main")
        if entry_point_python_script == None:
            print('package.json should have a key called {"main": "main.py"}')
            return

        self._add_to_gitignore(".build/")
        self._add_to_gitignore(".dist/")
        self._add_to_gitignore("*.spec")
        self._add_to_gitignore(".yppm_dist/")

        disk.create_a_folder("dist")

        terminal.run(f"""
        {self._hack_into_virtual_env_bash_command()}

        export PIP_BREAK_SYSTEM_PACKAGES=1
        {pip_path} install pyinstaller==5.13.0
        """)

        pyinstaller_path = self._get_virtual_env_program("pyinstaller", exact_match=True)
        if pyinstaller_path == None:
            pyinstaller_path = f"{python_path} -m pyinstaller"

        try:
            hidden_import_list = [one.split("=")[0] for one in package_object.get("dependencies") if not (one.startswith("git") or one.startswith("http"))]
            hidden_import_list = [f"--hidden-import {one}" for one in hidden_import_list]
            hidden_import_list_text = " ".join(hidden_import_list)
        except Exception as e:
            print(e)
            hidden_import_list_text = ""

        terminal.run(f"""
        {self._hack_into_virtual_env_bash_command()}

        {pyinstaller_path} {entry_point_python_script} {arguments} --noconfirm --onefile --add-data "./auto_everything:auto_everything" {hidden_import_list_text} --name {name}
         """)

    def build_with_nuitka(self, arguments: str = ""):
        self._create_virtual_env()
        python_path = self._get_virtual_env_python_excutable_path()
        pip_path = self._get_virtual_env_pip_path()

        package_object = self._get_package_json_object()

        name = package_object.get("name")
        if name == None:
            print('package.json should have a key called {"name": "your_app_name"}')
            return

        entry_point_python_script = package_object.get("main")
        if entry_point_python_script == None:
            print('package.json should have a key called {"main": "main.py"}')
            return

        self._add_to_gitignore("*.dist/")
        self._add_to_gitignore("*.build/")

        disk.create_a_folder("dist")

        terminal.run(f"""
        {self._hack_into_virtual_env_bash_command()}

        export PIP_BREAK_SYSTEM_PACKAGES=1
        {pip_path} install nuitka==1.8.4
        {pip_path} install patchelf==0.17.2.1

        {python_path} -m nuitka {arguments} --static-libpython=yes --standalone --follow-imports {entry_point_python_script} -o {name}
         """)

    def clean(self):
        disk.delete_a_folder(self.virtual_env_folder)
        disk.delete_a_folder("./build")
        disk.delete_a_folder("./dist")
        terminal.run(f"""
        rm -fr *.spec
                     """)

    # def env(self):
    #     print(f""" Please run:\n\n{self.env_activate_file_path} """.strip())

    def about(self):
        terminal.run("clear")
        print("""
YPPM: Yingshaoxo Python Package Manager.

(yingshaoxo is my name)
        """.strip())


try:
    py.fire2(Tools)
except Exception as e:
    print(e)
    pass



from typing import Callable

from .app_store_objects import *
from auto_everything.database import Database_Of_Yingshaoxo


def _search_function(self: Any, item_filter: Any, page_number:int|None=None, page_size:int|None=None, start_from:int=0, reverse:bool=False) -> list[dict[str, Any]]:
    search_temp_dict = {}
    search_temp_dict["_raw_search_counting"] = 0
    search_temp_dict["_search_counting"] = 0
    if (page_number!=None and page_size != None and start_from != None):
        search_temp_dict["_real_start"] = page_number * page_size
        search_temp_dict["_real_end"] = search_temp_dict["_real_start"] + page_size

    item_dict = item_filter.to_dict()

    def one_row_dict_filter(a_dict_: dict[str, Any]):
        search_temp_dict["_raw_search_counting"] += 1

        if (page_number!=None and page_size != None and start_from != None):
            if search_temp_dict["_raw_search_counting"] < start_from:
                return None

        result = True
        for key, value in item_dict.items():
            if value == None:
                # ignore None value because it is not defined
                continue
            if key not in a_dict_.keys():
                result = False
                break
            else:
                value2 = a_dict_.get(key)
                if value == value2:
                    continue
                else:
                    result = False
                    break

        final_result = None
        if result == True:
            search_temp_dict["_search_counting"] += 1
            final_result = a_dict_
        else:
            final_result = None

        if (page_number!=None and page_size != None and start_from != None):
            if search_temp_dict["_search_counting"] <= search_temp_dict["_real_start"]:
                return None
            if search_temp_dict["_search_counting"] > search_temp_dict["_real_end"]:
                return None

        return final_result

    return self.database_of_yingshaoxo.search(one_row_dict_handler=one_row_dict_filter)


def _raw_search_function(self: Any, one_row_json_string_handler: Callable[[str], dict[str, Any] | None], page_number:int|None=None, page_size:int|None=None, start_from:int=0, reverse:bool=False):
    search_temp_dict = {}
    search_temp_dict["_raw_search_counting"] = 0
    search_temp_dict["_search_counting"] = 0
    if (page_number!=None and page_size != None and start_from != None):
        search_temp_dict["_real_start"] = page_number * page_size
        search_temp_dict["_real_end"] = search_temp_dict["_real_start"] + page_size

    def new_one_row_json_string_handler(a_json_string: str):
        search_temp_dict["_raw_search_counting"] += 1

        if (page_number!=None and page_size != None and start_from != None):
            if search_temp_dict["_raw_search_counting"] < start_from:
                return None

        result = one_row_json_string_handler(a_json_string)

        if result != None:
            search_temp_dict["_search_counting"] += 1

        if (page_number!=None and page_size != None and start_from != None):
            if search_temp_dict["_search_counting"] <= search_temp_dict["_real_start"]:
                return None
            if search_temp_dict["_search_counting"] > search_temp_dict["_real_end"]:
                return None

        return result

    return list(self.database_of_yingshaoxo.raw_search(one_row_json_string_handler=new_one_row_json_string_handler))


def _delete(self, item_filter: Any) -> None:
    item_dict = item_filter.to_dict()
    def one_row_dict_filter(a_dict_: dict[str, Any]):
        result = True
        for key, value in item_dict.items():
            if value == None:
                # ignore None value because it is not defined
                continue
            if key not in a_dict_.keys():
                result = False
                break
            else:
                value2 = a_dict_.get(key)
                if value == value2:
                    continue
                else:
                    result = False
                    break
        return result
    self.database_of_yingshaoxo.delete(one_row_dict_filter=one_row_dict_filter)


def _update(self, old_item_filter: Any, new_item: Any):
    item_dict = old_item_filter.to_dict()
    def one_row_dict_handler(a_dict_: dict[str, Any]):
        result = True
        for key, value in item_dict.items():
            if value == None:
                # ignore None value because it is not defined
                continue
            if key not in a_dict_.keys():
                result = False
                break
            else:
                value2 = a_dict_.get(key)
                if value == value2:
                    continue
                else:
                    result = False
                    break
        if result == True:
            new_object = {
                key:value for key, value
                in new_item.to_dict().items()
                if value != None
            }
            a_dict_.update(new_object)
            return a_dict_
        else:
            return None
    self.database_of_yingshaoxo.update(one_row_dict_handler=one_row_dict_handler)


class Yingshaoxo_Database_An_App:
    def __init__(self, database_base_folder: str, use_sqlite: bool = False, global_multiprocessing_shared_dict: Any | None = None) -> None:
        self.database_of_yingshaoxo = Database_Of_Yingshaoxo(database_name="An_App", database_base_folder=database_base_folder, use_sqlite=use_sqlite, global_multiprocessing_shared_dict=global_multiprocessing_shared_dict)

    def add(self, item: An_App):
        return self.database_of_yingshaoxo.add(data=item.to_dict())

    def search(self, item_filter: An_App, page_number:int|None=None, page_size:int|None=None, start_from:int=0, reverse:bool=False) -> list[An_App]:
        return [An_App().from_dict(one) for one in _search_function(self=self, item_filter=item_filter, page_number=page_number, page_size=page_size, start_from=start_from, reverse=reverse)]

    def raw_search(self, one_row_json_string_handler: Callable[[str], dict[str, Any] | None], page_number:int|None=None, page_size:int|None=None, start_from:int=0, reverse:bool=False) -> list[An_App]:
        '''
        one_row_json_string_handler: a_function to handle search process. If it returns None, we'll ignore it, otherwise, we'll add the return value into the result list.
        '''
        return [An_App().from_dict(one) for one in _raw_search_function(self=self, one_row_json_string_handler=one_row_json_string_handler, page_number=page_number, page_size=page_size, start_from=start_from, reverse=reverse)]

    def delete(self, item_filter: An_App):
        return _delete(self=self, item_filter=item_filter)

    def update(self, old_item_filter: An_App, new_item: An_App):
        return _update(self=self, old_item_filter=old_item_filter, new_item=new_item)


class Yingshaoxo_Database_Excutor_app_store:
    def __init__(self, database_base_folder: str, use_sqlite: bool = False, global_multiprocessing_shared_dict: Any | None = None):
        self._database_base_folder = database_base_folder
        self.An_App = Yingshaoxo_Database_An_App(database_base_folder=self._database_base_folder, use_sqlite=use_sqlite, global_multiprocessing_shared_dict=global_multiprocessing_shared_dict)


if __name__ == "__main__":
    database_excutor = Yingshaoxo_Database_Excutor_app_store(database_base_folder="/home/yingshaoxo/CS/auto_everything/example/database/yingshaoxo_database")


from typing import Any, Callable, Iterator
from datetime import datetime
import json
import re


class MongoDB:
    """
    MONGO_DB_URL is something like: "mongodb://yingshaoxo:yingshaoxo@127.0.0.1:27017/"

    #https://www.w3schools.com/python/python_mongodb_create_collection.asp
    """
    def __init__(self, url: str):
        from pymongo import MongoClient
        from auto_everything.disk import Disk
        from auto_everything.io import IO
        self.client = MongoClient(url) #type: ignore
        self._disk = Disk()
        self._io = IO()

    @staticmethod
    def _get_mongodb_client_by_giving_arguments(host: str, port: str, user: str, password: str): #type: ignore
        return MongoDB(f"mongodb://{user}:{password}@{host}:{port}") #type: ignore

    def delete_a_database(self, database_name: str):
        self.client.drop_database(database_name) #type: ignore

    def list_database(self): #type: ignore
        return self.client.list_database_names() #type: ignore

    def get_database(self, database_name: str): #type: ignore
        return self.client.get_database(database_name) #type: ignore

    def insert_or_update_item(self, database_name: str | None, collection_name: str | None, filter: dict[str, Any] | None, data: dict[str, Any]): #type: ignore
        pass

    def get_item(self, database_name: str | None, collection_name: str | None, filter: dict[str, Any] | None) -> dict[str, Any]: #type: ignore
        pass

    def get_item_list(self, database_name: str | None, collection_name: str | None, filter: dict[str, Any] | None) -> list[dict[str, Any]]: #type: ignore
        pass

    def delete_item(self, database_name: str | None, collection_name: str | None, filter: dict[str, Any] | None) -> dict[str, Any]: #type: ignore
        pass

    def clear_mongodb(self, database_name: str|None = None, collection_name: str|None = None):
        database_name_list: list[str] = self.list_database()
        for a_database_name in database_name_list:
            if a_database_name in ["admin", "local", "config"]:
                continue

            if database_name == None and collection_name == None:
                self.client.drop_database(a_database_name) #type: ignore
            else:
                if database_name != None:
                    if database_name == a_database_name:
                        if collection_name != None:
                            self.client.get_database(name=a_database_name).get_collection(name=collection_name).delete_many({}) # type: ignore
                        else:
                            self.client.drop_database(name_or_database=a_database_name) # type: ignore

    def backup_collection(self, database_name: str, collection_name: str, json_file_saving_path: str | None=None) -> list[dict[Any, Any]]:
        database_name_list: list[str] = self.list_database()
        if database_name not in database_name_list:
            raise Exception(f"database_name '{database_name}' not exists")
        the_database = self.get_database(database_name=database_name) #type: ignore
        if collection_name not in the_database.list_collection_names():
            raise Exception(f"collection_name '{collection_name}' not exists")
        the_collection = the_database.get_collection(name=collection_name) #type: ignore

        object_list: list[Any] = []
        for one_object in the_collection.find(): #type: ignore
            obj = dict(one_object) #type: ignore
            del obj["_id"]
            object_list.append(obj)

        if json_file_saving_path != None:
            if not json_file_saving_path.endswith(".json"):
                raise Exception(f"the json_file_saving_path you give me should be ended with '.json' other than '{json_file_saving_path}'")
            json_string_of_an_collection = json.dumps(object_list, indent=4)
            self._io.write(json_file_saving_path, json_string_of_an_collection)

        return object_list

    def recover_collection(self, database_name: str, collection_name: str, object_list: list[Any] | None = None, json_file_saving_path: str | None=None):
        if json_file_saving_path != None:
            if not self._disk.exists(json_file_saving_path):
                raise Exception(f"The json_file_saving_path you gave me does not exists: {json_file_saving_path}")
            json_text =  self._io.read(json_file_saving_path)
            object_list = json.loads(json_text)

        self.clear_mongodb(database_name=database_name, collection_name=collection_name)
        self.get_database(database_name=database_name).get_collection(name=collection_name).insert_many(object_list) #type: ignore

    # def backup_mongodb_safely(self, backup_folder_path: str, use_time_as_folder_name: bool=True):
    #     # save one object as a json_string per line in a txt file
    #     pass

    def backup_mongodb(self, backup_folder_path: str, use_time_as_sub_folder_name: bool=True):
        """
        backup_folder_path: str
            the backup folder
        use_time_as_sub_folder_name: bool
            if it is true:
                it saves files into this structure:
                    ── 2023-03-31_09-30
                    │   ├── __mongodb_info__.json
                    │   │   database_name_1
                    │   │   ├── collection_name_1.json
                    │   │   ├── collection_name_2.json
                    │   │   database_name_2
                    │   │   ├── collection_name_1.json
                    │   │   ├── collection_name_2.json
            else:
                it saves files into this structure:
                    ├── __mongodb_info__.json
                    │   database_name_1
                    │   ├── collection_name_1.json
                    │   ├── collection_name_2.json
                    │   database_name_2
                    │   ├── collection_name_1.json
                    │   ├── collection_name_2.json
        """
        if not self._disk.exists(backup_folder_path):
            self._disk.create_a_folder(backup_folder_path)
        if not self._disk.exists(backup_folder_path):
            raise Exception(f"You should give me a valid backup_folder_path than '{backup_folder_path}'")

        if use_time_as_sub_folder_name == True:
            now = datetime.now()
            today_string = now.strftime(r"%Y-%m-%d_%H-%M")
            backup_folder_path = self._disk.join_paths(backup_folder_path, today_string)

        database_name_list: list[str] = self.list_database()
        for database_name in database_name_list:
            if database_name in ["admin", "local", "config"]:
                continue

            a_database = self.get_database(database_name=database_name) #type: ignore

            database_folder_path = self._disk.join_paths(backup_folder_path, database_name)
            self._disk.create_a_folder(database_folder_path)

            collection_name_list = a_database.list_collection_names()
            for collection_name in collection_name_list:
                collection = a_database.get_collection(name=collection_name) #type: ignore
                object_list: list[Any] = []
                for one_object in collection.find(): #type: ignore
                    obj = dict(one_object) #type: ignore
                    del obj["_id"]
                    object_list.append(obj)
                try:
                    json_string_of_an_collection = json.dumps(object_list, indent=4)
                except Exception as e:
                    print(e)
                    continue

                collection_json_file_path = self._disk.join_paths(database_folder_path, collection_name + ".json")
                self._io.write(collection_json_file_path, json_string_of_an_collection)

        special_file_path = self._disk.join_paths(backup_folder_path, "__mongodb_info__.json")
        file_name_tree = self._disk.fake_folder_backup(backup_folder=backup_folder_path)
        self._io.write(special_file_path, json.dumps({
            "db_type": "mongodb",
            "backup_tool": "python, auto_everything",
            "author": "yingshaoxo@gmail.com",
            "file_tree": file_name_tree
        }, indent=4))

    def recover_mongodb(self, backup_folder_path: str, use_time_as_sub_folder_name: bool=True):
        """
        backup_folder_path: str
            the backup folder
        use_time_as_sub_folder_name: bool
            if it is true:
                it loads files from this structure:
                    ── 2023-03-31_09-30
                    │   ├── __mongodb_info__.json
                    │   │   database_name_1
                    │   │   ├── collection_name_1.json
                    │   │   ├── collection_name_2.json
                    │   │   database_name_2
                    │   │   ├── collection_name_1.json
                    │   │   ├── collection_name_2.json
            else:
                it loads files from this structure:
                    ├── __mongodb_info__.json
                    │   database_name_1
                    │   ├── collection_name_1.json
                    │   ├── collection_name_2.json
                    │   database_name_2
                    │   ├── collection_name_1.json
                    │   ├── collection_name_2.json
        """
        if (use_time_as_sub_folder_name == True):
            date_folders = [one for one in self._disk.get_folder_and_files(folder=backup_folder_path, recursive=False) if one.is_folder == True]
            date_folders.sort(key=lambda one: one.name)
            if len(date_folders) > 0:
                backup_folder_path = date_folders[-1].path
            else:
                raise Exception(f"This folder does not contain any data: {backup_folder_path}")

        if not self._disk.exists(backup_folder_path):
            raise Exception(f"You should give me a valid backup_folder_path than '{backup_folder_path}'")
        if not self._disk.exists(self._disk.join_paths(backup_folder_path, "__mongodb_info__.json")):
            raise Exception(f"You should give me a valid backup_folder_path than '{backup_folder_path}'")

        self.clear_mongodb()

        database_folder_list = [one for one in self._disk.get_folder_and_files(folder=backup_folder_path, recursive=False) if one.is_folder]
        for database_folder in database_folder_list:
            database_name = database_folder.name
            a_database = self.get_database(database_name=database_name) #type: ignore

            collection_json_list = self._disk.get_folder_and_files(folder=self._disk.join_paths(backup_folder_path, database_name), recursive=False, type_limiter=[".json"])
            for collection_json_file in collection_json_list:
                collection_name = collection_json_file.name[:-len(".json")]
                json_content = self._io.read(collection_json_file.path)
                data_ = json.loads(json_content)
                for one in data_:
                    a_database.get_collection(collection_name).insert_one(one) #type: ignore


class Redis:
    """
    Redis_DB_URL is something like: "redis://username:password@127.0.0.1:6379"
    """
    def __init__(self, redis_URL: str, database_name: str) -> None:
        import redis
        from auto_everything.disk import Disk
        from auto_everything.io import IO
        self._io = IO()
        self._disk = Disk()

        host, port, username, password = self._parse_redis_url(redis_URL=redis_URL)

        self.redis = redis.Redis(host=host, port=int(port), db=0, username=username, password=password)
        self.database_name = database_name

    @staticmethod
    def _parse_redis_url(redis_URL: str) -> tuple[str, str, str | None, str | None]:
        if not redis_URL.startswith("redis://"):
            raise Exception(f"the redis_URL is something like this: redis://127.0.0.1:6379")
        username = None
        password = None
        host = None
        port = None
        if "@" in redis_URL:
            splits = redis_URL.split("@")
            first_part = splits[0].split("://")[1]
            second_part = splits[1]

            splits = first_part.split(":")
            username = splits[0]
            password = splits[1]

            splits = second_part.split(":")
            host = splits[0]
            port = splits[1].split("/")[0]
        else:
            'redis://127.0.0.1:6379/0'
            ip_and_port = redis_URL.split("://")[1].split("/")[0]
            splits = ip_and_port.split(":")
            host = splits[0]
            port = splits[1]
        return host, port, username, password

    def _get_final_key_with_database_name_as_prefix(self, key: str) -> str:
        return f"{self.database_name}.{key}"

    def get(self, key: str) -> str | None:
        data = self.redis.get(
            self._get_final_key_with_database_name_as_prefix(key=key)
        )
        if data is None:
            return None
        else:
            return data.decode('utf-8')

    def set(self, key: str, value: str, expire_time_in_seconds: int | None = None) -> bool:
        result = self.redis.set(
            self._get_final_key_with_database_name_as_prefix(key=key),
            value,
            ex=expire_time_in_seconds
        )
        if result is None:
            return False
        else:
            return result

    def delete(self, key: str) -> bool:
        result = self.redis.delete(
            self._get_final_key_with_database_name_as_prefix(key=key)
        )
        if result == 0:
            return False
        else:
            return True

    def delete_all(self):
        for key in self.redis.keys(
                self._get_final_key_with_database_name_as_prefix(key="*")
            ):
            self.redis.delete(key)

    def get_remaining_time_of_a_key(self, key: str) -> int:
        """
        The command returns -2 if the key does not exist.
        The command returns -1 if the key exists but has no associated expire.
        """
        return self.redis.ttl(
            self._get_final_key_with_database_name_as_prefix(key=key)
        )

    def backup_redis(self, json_file_saving_path: str | None = None) -> list[dict[str, str]]:
        value_list: list[dict[str, str]] = []
        for key in self.redis.keys(
                self._get_final_key_with_database_name_as_prefix(key="*")
            ):
            value = self.redis.get(key)
            if value != None:
                object = {
                    "key": key.decode("utf-8"),
                    "value": value.decode("utf-8"),
                }
                ttl = self.get_remaining_time_of_a_key(key=key.decode("utf-8"))
                if ttl >= 1:
                    object["ttl"] = str(ttl)
                value_list.append(object)

        if json_file_saving_path != None:
            if not json_file_saving_path.endswith(".json"):
                raise Exception(f"the json_file_saving_path you give me should be ended with '.json' other than '{json_file_saving_path}'")
            json_string_of_an_collection = json.dumps(value_list, indent=4)
            self._io.write(json_file_saving_path, json_string_of_an_collection)

        return value_list

    def recover_redis(self, object_list: list[Any] | None = None, json_file_saving_path: str | None=None):
        if json_file_saving_path != None:
            if not self._disk.exists(json_file_saving_path):
                raise Exception(f"The json_file_saving_path you gave me does not exists: {json_file_saving_path}")
            json_text =  self._io.read(json_file_saving_path)
            object_list = json.loads(json_text)

        self.delete_all()
        for one in object_list: #type: ignore
            key, value, ttl = one.get("key"), one.get("value"), one.get("ttl")
            if ttl == None:
                self.set(key=key, value=value)
            else:
                self.set(key=key, value=value, expire_time_in_seconds=int(ttl))


class Database_Of_Yingshaoxo:
    """
                        r   r+   w   w+   a   a+
    ------------------|--------------------------
    read              | +   +        +        +
    write             |     +    +   +    +   +
    write after seek  |     +    +   +
    create            |          +   +    +   +
    truncate          |          +   +
    position at start | +   +    +   +
    position at end   |                   +   +
    """
    def __init__(self, database_name: str, database_base_folder: str = "./yingshaoxo_database", use_sqlite: bool = False) -> None:
        from auto_everything.disk import Disk
        from auto_everything.io import IO
        import json
        import subprocess
        import os
        self._disk = Disk()
        self._io = IO()
        self._json = json
        self._subprocess = subprocess
        self._os = os

        self.database_base_folder = database_base_folder
        if (not self._disk.exists(self.database_base_folder)):
            self._disk.create_a_folder(self.database_base_folder)

        self.use_sqlite = use_sqlite

        if self.use_sqlite == False:
            self.database_txt_file_path = self._disk.join_paths(self.database_base_folder, f"{database_name}.txt")
            if (not self._disk.exists(self.database_txt_file_path)):
                self._io.write(self.database_txt_file_path, "")
        else:
            import sqlite3 as sqlite3

            self._SQL_DATA_FILE = self._disk.join_paths(self.database_base_folder, f"{database_name}.db")
            """
            if sqlite3.threadsafety == 3:
                check_same_thread = False
            else:
                check_same_thread = True
            """
            check_same_thread = False
            self.sql_connection = sqlite3.connect(self._SQL_DATA_FILE, check_same_thread=check_same_thread)

            def regular_expression(expr: str, item: Any):
                reg = re.compile(expr, flags=re.DOTALL)
                return reg.search(item) is not None

            self.sql_connection.create_function(
                "REGEXP", 2, regular_expression
            )  # 2 here means two parameters. REGEXP is a fixed value

            self._sql_table_name = database_name.replace(" ", "_")
            self.sql_cursor = self.sql_connection.cursor()
            self.sql_cursor.execute(
                f"""CREATE TABLE IF NOT EXISTS {self._sql_table_name}
                        (value TEXT)"""
            )

    def add(self, data: dict[str, Any]):
        json_string = self._json.dumps(data, sort_keys=True).strip()

        if self.use_sqlite:
            self.sql_cursor.execute(
                f"INSERT INTO {self._sql_table_name} (value) VALUES (?)", (json_string, )
            )
            self.sql_connection.commit()
        else:
            with open(self.database_txt_file_path, "a+", encoding="utf-8", errors="ignore") as file_stream:
                file_stream.seek(0, self._os.SEEK_END)
                file_stream.write(json_string + "\n")

    def raw_search(self, one_row_json_string_handler: Callable[[str], dict[str, Any] | None]) -> Iterator[dict[str, Any]]:
        """
        one_row_json_string_handler: a_function to handle search process. If it returns None, we'll ignore it, otherwise, we'll add the return value into the result list.
        ```
            def one_row_json_string_handler(item: str) -> dict[str, Any] | None:
                return json.loads(item)
                #return None
        ```
        """
        if self.use_sqlite:
            for row in self.sql_cursor.execute(
                f"SELECT * FROM {self._sql_table_name}"
            ):
                result = one_row_json_string_handler(row[0])
                if (result != None):
                    yield result
        else:
            with open(self.database_txt_file_path, "r") as file_stream:
                previous_position = None
                while True:
                    current_position = file_stream.tell()
                    line = file_stream.readline()
                    if previous_position == current_position:
                        # reach the end
                        break
                    previous_position = current_position
                    if (line.strip() == ""):
                        # ignore empty line
                        continue

                    if (line.startswith('#')):
                        # ignore deleted line
                        continue

                    result = one_row_json_string_handler(line)
                    if (result != None):
                        yield result

    def search(self, one_row_dict_handler: Callable[[dict[str, Any]], dict[str, Any] | None]) -> list[dict[str, Any]]:
        """
        one_row_dict_handler: a_function to handle search process. If it returns None, we'll ignore it, otherwise, we'll add the return value into the result list.
        ```
            def one_row_dict_handler(item: dict[str, Any]) -> dict[str, Any] | None:
                return item
                #return None
        ```
        """
        if self.use_sqlite:
            result_list = []
            for row in self.sql_cursor.execute(
                f"SELECT * FROM {self._sql_table_name}"
            ):
                json_dict = self._json.loads(row[0])
                result = one_row_dict_handler(json_dict)
                if (result != None):
                    result_list.append(result)
            return result_list
        else:
            result_list = []
            with open(self.database_txt_file_path, "r") as file_stream:
                previous_position = None
                while True:
                    current_position = file_stream.tell()
                    line = file_stream.readline()
                    if previous_position == current_position:
                        # reach the end
                        break
                    previous_position = current_position
                    if (line.strip() == ""):
                        # ignore empty line
                        continue

                    if (line.startswith('#')):
                        # ignore deleted line
                        continue

                    json_dict = self._json.loads(line)

                    result = one_row_dict_handler(json_dict)
                    if (result != None):
                        result_list.append(result)
            return result_list

    # def reverse_search(self):
    #     #https://stackoverflow.com/a/23646049/8667243
    #     #reverse search will speed up the search process in most of the cases
    #     pass

    def raw_delete(self, one_row_json_string_filter: Callable[[str], bool]):
        """
        one_row_json_string_filter: a_function to handle deletion process. If it returns False, we'll ignore it, otherwise, if it is True, we'll delete that row of data.
        ```
            def one_row_json_string_filter(item_string: str) -> bool:
                item = json.loads(item_string)
                return False
                #return True
        ```
        """
        if self.use_sqlite:
            need_to_get_deleted_value_list = []
            for row in self.sql_cursor.execute(
                f"SELECT * FROM {self._sql_table_name}"
            ):
                result = one_row_json_string_filter(row[0])
                if (result == True):
                    need_to_get_deleted_value_list.append(row[0])

            for value in need_to_get_deleted_value_list:
                self.sql_cursor.execute(
                    f"DELETE FROM {self._sql_table_name} WHERE value=?", (value,)
                )

            self.sql_connection.commit()
        else:
            with open(self.database_txt_file_path, "r+") as file_stream:
                end_detection_counting = 1
                old_position_pair = None
                while True:
                    previous_position = file_stream.tell()
                    line = file_stream.readline()
                    current_position = file_stream.tell()

                    new_position_pair = (previous_position, current_position)
                    if old_position_pair == new_position_pair:
                        end_detection_counting += 1
                    else:
                        old_position_pair = new_position_pair
                    if end_detection_counting >= 3:
                        # We could make sure it is the end of the file
                        old_position_pair = None
                        break

                    if (line.strip() == ""):
                        # ignore empty line
                        continue

                    if (line.startswith('#')):
                        # ignore deleted line
                        continue

                    result = one_row_json_string_filter(line)
                    if (result == True):
                        # replace the first character of the line with '#' symbol
                        file_stream.seek(previous_position)
                        file_stream.write("#"+line[1:])

    def delete(self, one_row_dict_filter: Callable[[dict[str, Any]], bool]):
        """
        one_row_dict_filter: a_function to handle deletion process. If it returns False, we'll ignore it, otherwise, if it is True, we'll delete that row of data.
        ```
            def one_row_dict_filter(item: dict[str, Any]) -> bool:
                return False
                #return True
        ```
        """
        if self.use_sqlite:
            need_to_get_deleted_value_list = []
            for row in self.sql_cursor.execute(
                f"SELECT * FROM {self._sql_table_name}"
            ):
                result = one_row_dict_filter(json.loads(row[0]))
                if (result == True):
                    need_to_get_deleted_value_list.append(row[0])

            for value in need_to_get_deleted_value_list:
                self.sql_cursor.execute(
                    f"DELETE FROM {self._sql_table_name} WHERE value=?", (value,)
                )

            self.sql_connection.commit()
        else:
            with open(self.database_txt_file_path, "r+") as file_stream:
                end_detection_counting = 1
                old_position_pair = None
                while True:
                    previous_position = file_stream.tell()
                    line = file_stream.readline()
                    current_position = file_stream.tell()

                    new_position_pair = (previous_position, current_position)
                    #print(new_position_pair)
                    if old_position_pair == new_position_pair:
                        end_detection_counting += 1
                    else:
                        old_position_pair = new_position_pair
                    if end_detection_counting >= 3:
                        # We could make sure it is the end of the file
                        old_position_pair = None
                        break

                    if (line.strip() == ""):
                        # ignore empty line
                        continue

                    if (line.startswith('#')):
                        # ignore deleted line
                        continue

                    json_dict = self._json.loads(line)

                    result = one_row_dict_filter(json_dict)
                    #print(result)
                    if (result == True):
                        # replace the first character of the line with '#' symbol
                        file_stream.seek(previous_position)
                        file_stream.write("#"+line[1:])

    def raw_update(self, one_row_json_string_handler: Callable[[str], dict[str, Any] | None]):
        """
        one_row_json_string_handler: a_function to handle update process. If it returns None, we'll ignore it, otherwise, we'll update the old value with the new value the handler function returns.
        ```
            def one_row_json_string_handler(item_string: str) -> dict[str, Any] | None:
                return json,loads(item_string)
                #return None
        ```
        """
        if self.use_sqlite:
            need_to_get_updated_value_list = []

            for row in self.sql_cursor.execute(
                f"SELECT * FROM {self._sql_table_name}"

            ):
                result = one_row_json_string_handler(row[0])
                if (result != None):
                    need_to_get_updated_value_list.append([row[0], json.dumps(result)])

            for item in need_to_get_updated_value_list:
                self.sql_cursor.execute(
                    f"UPDATE {self._sql_table_name} SET value=? WHERE value=?", (item[1], item[0], )
                )

            self.sql_connection.commit()
        else:
            new_record_list = []

            def one_row_dict_filter(old_value: str) -> bool:
                result = one_row_json_string_handler(old_value)
                if result == None:
                    return False
                else:
                    new_record_list.append(result)
                    return True
            self.raw_delete(one_row_json_string_filter=one_row_dict_filter)

            for one in new_record_list:
                self.add(one)

    def update(self, one_row_dict_handler: Callable[[dict[str, Any]], dict[str, Any] | None]):
        """
        one_row_dict_handler: a_function to handle update process. If it returns None, we'll ignore it, otherwise, we'll update the old value with the new value the handler function returns.
        ```
            def one_row_dict_handler(item: dict[str, Any]) -> dict[str, Any] | None:
                return item
                #return None
        ```
        """
        if self.use_sqlite:
            need_to_get_updated_value_list = []

            for row in self.sql_cursor.execute(
                f"SELECT * FROM {self._sql_table_name}"

            ):
                result = one_row_dict_handler(json.loads(row[0]))
                if (result != None):
                    need_to_get_updated_value_list.append([row[0], json.dumps(result)])

            for item in need_to_get_updated_value_list:
                self.sql_cursor.execute(
                    f"UPDATE {self._sql_table_name} SET value=? WHERE value=?", (item[1], item[0], )
                )

            self.sql_connection.commit()
        else:
            new_record_list = []

            def one_row_dict_filter(old_value: dict[str, Any]) -> bool:
                result = one_row_dict_handler(old_value)
                if result == None:
                    return False
                else:
                    new_record_list.append(result)
                    return True
            self.delete(one_row_dict_filter=one_row_dict_filter)

            for one in new_record_list:
                self.add(one)

    def refactor_database(self):
        if self.use_sqlite:
            return

        text = self._io.read(self.database_txt_file_path)
        lines = []
        for line in [line.strip() for line in text.split("\n") if line.strip() != ""]:
            if not line.startswith('#'):
                lines.append(line)
        self._io.write(self.database_txt_file_path, "\n".join(lines)+"\n")

    def clear_database(self):
        if self.use_sqlite:
            self.sql_cursor.execute(
                f"DELETE FROM {self._sql_table_name}"
            )
            self.sql_connection.commit()
            return

        self._io.write(self.database_txt_file_path, "")

    @staticmethod
    def generate_code_from_yrpc_protocol(which_language: str, input_folder: str, input_files: list[str], output_folder: str = "src/generated_yrpc"):
        """
        which_language: python. #dart, typescript, go, kotlin, rust and so on

        input_folder: where protobuff files was located

        input_files: it is a list, like ["english.proto", "pornhub.proto"]

        output_folder: where those generated code file was located
        """
        from auto_everything.disk import Disk
        from auto_everything.io import IO
        from auto_everything.develop import YRPC
        import re
        disk_ = Disk()
        io_ = IO()
        yrpc_ = YRPC()

        def _convert_yrpc_code_into_yingshaoxo_database_python_rpc_code(identity_name: str, source_code: str) -> str:
            arguments_defination_tree, rpc_dict = yrpc_.get_information_from_yrpc_protocol_code(source_code=source_code)

            arguments_name_list = [one for one in arguments_defination_tree.keys() if arguments_defination_tree[one]['**type**'] == 'message']
            rpc_arguments_name_list: list[str] = []

            for function_name, parameter_info in rpc_dict.items():
                input_variable: str = parameter_info["input_variable"]
                output_variable: str = parameter_info["output_variable"]

                if " " in input_variable:
                    input_variable = re.split(r"\s+", input_variable)[1]
                if " " in output_variable:
                    output_variable = re.split(r"\s+", output_variable)[1]

                variable_list = list(set([input_variable, output_variable]))
                for variable_type in variable_list:
                    if variable_type not in rpc_arguments_name_list:
                        rpc_arguments_name_list.append(variable_type)

            data_model_name_list = []
            for each_one in arguments_name_list:
                if each_one not in rpc_arguments_name_list:
                    data_model_name_list.append(each_one)

            database_class_list: list[str] = []
            database_excutor_class_property_list: list[str] = []
            for variable_type in data_model_name_list:
                database_class_list.append(f"""
class Yingshaoxo_Database_{variable_type}:
    def __init__(self, database_base_folder: str, use_sqlite: bool = False) -> None:
        self.database_of_yingshaoxo = Database_Of_Yingshaoxo(database_name="{variable_type}", database_base_folder=database_base_folder, use_sqlite=use_sqlite)

    def add(self, item: {variable_type}):
        return self.database_of_yingshaoxo.add(data=item.to_dict())

    def search(self, item_filter: {variable_type}, page_number:int|None=None, page_size:int|None=None, start_from:int=0, reverse:bool=False) -> list[{variable_type}]:
        return [{variable_type}().from_dict(one) for one in _search_function(self=self, item_filter=item_filter, page_number=page_number, page_size=page_size, start_from=start_from, reverse=reverse)]

    def raw_search(self, one_row_json_string_handler: Callable[[str], dict[str, Any] | None], page_number:int|None=None, page_size:int|None=None, start_from:int=0, reverse:bool=False) -> list[{variable_type}]:
        '''
        one_row_json_string_handler: a_function to handle search process. If it returns None, we'll ignore it, otherwise, we'll add the return value into the result list.
        '''
        return [{variable_type}().from_dict(one) for one in _raw_search_function(self=self, one_row_json_string_handler=one_row_json_string_handler, page_number=page_number, page_size=page_size, start_from=start_from, reverse=reverse)]

    def delete(self, item_filter: {variable_type}):
        return _delete(self=self, item_filter=item_filter)

    def update(self, old_item_filter: {variable_type}, new_item: {variable_type}):
        return _update(self=self, old_item_filter=old_item_filter, new_item=new_item)
                """.rstrip().lstrip('\n'))

            for variable_type in data_model_name_list:
                database_excutor_class_property_list.append(f"""
        self.{variable_type} = Yingshaoxo_Database_{variable_type}(database_base_folder=self._database_base_folder, use_sqlite=use_sqlite)
                """.rstrip().lstrip('\n'))


            database_class_list_text = "\n\n\n".join(database_class_list)
            database_excutor_class_property_list_text = "\n".join(database_excutor_class_property_list)

            template_text = f"""
from typing import Callable

from .{identity_name}_objects import *
from auto_everything.database import Database_Of_Yingshaoxo


def _search_function(self: Any, item_filter: Any, page_number:int|None=None, page_size:int|None=None, start_from:int=0, reverse:bool=False) -> list[dict[str, Any]]:
    search_temp_dict = {{}}
    search_temp_dict["_raw_search_counting"] = 0
    search_temp_dict["_search_counting"] = 0
    if (page_number!=None and page_size != None and start_from != None):
        search_temp_dict["_real_start"] = page_number * page_size
        search_temp_dict["_real_end"] = search_temp_dict["_real_start"] + page_size

    item_dict = item_filter.to_dict()

    def one_row_dict_filter(a_dict_: dict[str, Any]):
        search_temp_dict["_raw_search_counting"] += 1

        if (page_number!=None and page_size != None and start_from != None):
            if search_temp_dict["_raw_search_counting"] < start_from:
                return None

        result = True
        for key, value in item_dict.items():
            if value == None:
                # ignore None value because it is not defined
                continue
            if key not in a_dict_.keys():
                result = False
                break
            else:
                value2 = a_dict_.get(key)
                if value == value2:
                    continue
                else:
                    result = False
                    break

        final_result = None
        if result == True:
            search_temp_dict["_search_counting"] += 1
            final_result = a_dict_
        else:
            final_result = None

        if (page_number!=None and page_size != None and start_from != None):
            if search_temp_dict["_search_counting"] <= search_temp_dict["_real_start"]:
                return None
            if search_temp_dict["_search_counting"] > search_temp_dict["_real_end"]:
                return None

        return final_result

    return self.database_of_yingshaoxo.search(one_row_dict_handler=one_row_dict_filter)


def _raw_search_function(self: Any, one_row_json_string_handler: Callable[[str], dict[str, Any] | None], page_number:int|None=None, page_size:int|None=None, start_from:int=0, reverse:bool=False):
    search_temp_dict = {{}}
    search_temp_dict["_raw_search_counting"] = 0
    search_temp_dict["_search_counting"] = 0
    if (page_number!=None and page_size != None and start_from != None):
        search_temp_dict["_real_start"] = page_number * page_size
        search_temp_dict["_real_end"] = search_temp_dict["_real_start"] + page_size

    def new_one_row_json_string_handler(a_json_string: str):
        search_temp_dict["_raw_search_counting"] += 1

        if (page_number!=None and page_size != None and start_from != None):
            if search_temp_dict["_raw_search_counting"] < start_from:
                return None

        result = one_row_json_string_handler(a_json_string)

        if result != None:
            search_temp_dict["_search_counting"] += 1

        if (page_number!=None and page_size != None and start_from != None):
            if search_temp_dict["_search_counting"] <= search_temp_dict["_real_start"]:
                return None
            if search_temp_dict["_search_counting"] > search_temp_dict["_real_end"]:
                return None

        return result

    return list(self.database_of_yingshaoxo.raw_search(one_row_json_string_handler=new_one_row_json_string_handler))


def _delete(self, item_filter: Any) -> None:
    item_dict = item_filter.to_dict()
    def one_row_dict_filter(a_dict_: dict[str, Any]):
        result = True
        for key, value in item_dict.items():
            if value == None:
                # ignore None value because it is not defined
                continue
            if key not in a_dict_.keys():
                result = False
                break
            else:
                value2 = a_dict_.get(key)
                if value == value2:
                    continue
                else:
                    result = False
                    break
        return result
    self.database_of_yingshaoxo.delete(one_row_dict_filter=one_row_dict_filter)


def _update(self, old_item_filter: Any, new_item: Any):
    item_dict = old_item_filter.to_dict()
    def one_row_dict_handler(a_dict_: dict[str, Any]):
        result = True
        for key, value in item_dict.items():
            if value == None:
                # ignore None value because it is not defined
                continue
            if key not in a_dict_.keys():
                result = False
                break
            else:
                value2 = a_dict_.get(key)
                if value == value2:
                    continue
                else:
                    result = False
                    break
        if result == True:
            new_object = {{
                key:value for key, value
                in new_item.to_dict().items()
                if value != None
            }}
            a_dict_.update(new_object)
            return a_dict_
        else:
            return None
    self.database_of_yingshaoxo.update(one_row_dict_handler=one_row_dict_handler)


{database_class_list_text}


class Yingshaoxo_Database_Excutor_{identity_name}:
    def __init__(self, database_base_folder: str, use_sqlite: bool = False):
        self._database_base_folder = database_base_folder
{database_excutor_class_property_list_text}


if __name__ == "__main__":
    database_excutor = Yingshaoxo_Database_Excutor_{identity_name}(database_base_folder="/home/yingshaoxo/CS/auto_everything/example/database/yingshaoxo_database")
            """.strip()

            return template_text

        input_folder = input_folder.rstrip("/")

        if not disk_.exists(input_folder):
            raise Exception(f"The input_forder '{input_folder}' does not exist!")

        if not disk_.is_directory(input_folder):
            raise Exception(f"The input_folder '{input_folder}' must be an directory.")

        if not disk_.is_directory(output_folder):
            disk_.create_a_folder(output_folder)

        files = disk_.get_files(input_folder, recursive=False, type_limiter=[".proto"])

        new_files:list[str] = []
        for file in files:
            if any([one for one in input_files if file.endswith("/"+one)]):
                new_files.append(file)
        files = new_files.copy()

        language_to_file_suffix_dict = {
            "python": ".py",
            # "dart": ".dart",
            # "typescript": ".ts",
            # "golang": ".go"
        }

        if which_language not in language_to_file_suffix_dict.keys():
            raise Exception(f"Sorry, we don't support '{which_language}' language.")

        if which_language == "python":
            init_file_for_python = disk_.join_paths(output_folder, "__init__.py")
            if not disk_.exists(init_file_for_python):
                io_.write(init_file_for_python, "")

        for file in files:
            filename,_ = disk_.get_stem_and_suffix_of_a_file(file)
            identity_name = filename
            if " " in identity_name:
                print(f"Sorry, protocol filename shoudn't have space inside: '{identity_name}'")
                exit()

            source_code = io_.read(file_path=file)

            objects_code = ""
            database_rpc_code = ""
            if which_language in ["python"]:
                target_objects_file_path = disk_.join_paths(output_folder, filename + "_objects" + language_to_file_suffix_dict[which_language])
                target_rpc_file_path = disk_.join_paths(output_folder, filename + "_yingshaoxo_database_rpc" + language_to_file_suffix_dict[which_language])

                if which_language == "python":
                    objects_code = yrpc_._convert_yrpc_code_into_python_objects_code(source_code=source_code)
                    database_rpc_code = _convert_yrpc_code_into_yingshaoxo_database_python_rpc_code(identity_name=identity_name, source_code=source_code)
                io_.write(file_path=target_objects_file_path, content=objects_code)
                io_.write(file_path=target_rpc_file_path, content=database_rpc_code)
            else:
                print(f"Sorry, we do not support this programming language: {which_language}")
                exit()


if __name__ == "__main__":
    pass

"""
import sqlalchemy

# Stupid MySQL can't be reasoned with
class MySQL:
    def __init__(self, host: str, port: str, user: str, password: str):
        self.engine = sqlalchemy.create_engine(f'mysql+pymysql://{user}:{password}@{host}:{port}')
        self._block_list = ['mysql', 'performance_schema', 'sys']

    def create_a_database(self, database_name: str):
        self.engine.execute(f"CREATE DATABASE {database_name}")
        #self.engine.execute(f"USE {database_name}")

    def delete_a_database(self, database_name: str):
        if (database_name not in self._block_list):
            self.engine.execute(f"DROP DATABASE IF EXISTS {database_name};")

    def list_database(self) -> str:
        existing_databases = self.engine.execute("SHOW DATABASES;")
        return [d[0] for d in existing_databases if d[0] not in self._block_list]

    def is_database_exists(self, database_name: str) -> bool:
        existing_databases = self.list_database()
        return database_name in existing_databases

if __name__ == "__main__":
    mySQL = MySQL(host="127.0.0.1", port="3306", user="root", password="yingshaoxo666")

    for d in mySQL.list_database():
        mySQL.delete_a_database(d)
    print(mySQL.list_database())

    mySQL.create_a_database("test")
    print(mySQL.list_database())

    for d in mySQL.list_database():
        mySQL.delete_a_database(d)
    print(mySQL.list_database())
"""



from typing import Any
import re


class Language():
    def compare_two_sentences(self, sentence1: str, sentence2: str) -> float:
        """
        return similarity, from `0.0` to `1.0`, 1 means equal, 0 means no relate.

        Parameters
        ----------
        sentence1: string

        sentence2: string
        """
        from difflib import SequenceMatcher
        ratio = SequenceMatcher(None, sentence1, sentence2).ratio()
        return ratio

    def seperate_text_to_segments(self, text: str, ignore_space: bool = True) -> list[dict[str, Any]]:
        """
        It returns a list of dict.
        For example:
            [
                {
                    "is_punctuation_or_space": true,
                    "text": "?",
                },
                {
                    "is_punctuation_or_space": false,
                    "text": "Yes",
                },
            ]
        """
        if ignore_space == True:
            all_result = re.findall(r"((?:\w* *)*)(\W*)", text, flags=re.MULTILINE)
        else:
            all_result = re.findall(r"(\w*)(\W*)", text, flags=re.MULTILINE)
        final_result = []
        for one in all_result:
            words = one[0]
            puctuation = one[1]
            if len(words) > 0:
                item = {
                    "is_punctuation_or_space": False,
                    "text": words
                }
                final_result.append(item)
            if len(puctuation) > 0:
                item = {
                    "is_punctuation_or_space": True,
                    "text": puctuation
                }
                final_result.append(item)
        return final_result
    

class English(Language):
    """
    This is a class for handling English
    """

    def __init__(self):
        from summa.summarizer import summarize
        from summa.keywords import keywords
        self._keywords = keywords
        self._summarize = summarize

    def text_to_speech(self, text):
        """
        speak accourding to text

        Parameters
        ----------
        text: string
        """
        import pyttsx3 as _pyttsx3
        self._speak_engine = _pyttsx3.init()
        self._speak_engine.setProperty('rate', 180)  # 180 words per minute
        self._speak_engine.setProperty('volume', 0.8)

        text = text.strip()
        self._speak_engine.say(text)
        self._speak_engine.runAndWait()
    
    def get_key_words(self, input_text: str) -> list[str]:
        return self._keywords(input_text) #type: ignore

    def get_sentences_from_text(self, text) -> list[str]:
        """
        return a list of sentences

        Parameters
        ----------
        text: string
        """
        from nltk import tokenize
        return tokenize.sent_tokenize(text)

    def get_key_sentences_from_text(self, text: str, reduce_ratio: float = 0.5, return_list: bool = True) -> list[str] | str:
        """
        return a list of key sentence

        Parameters
        ----------
        text: string
        reduce_ratio: float

        """
        if return_list:
            return self._summarize(text, split=True, ratio=reduce_ratio) #type: ignore
        else:
            return self._summarize(text, split=False, ratio=reduce_ratio) #type: ignore


class Chinese(Language):
    """
    This is a class for handling Chinese
    """

    def __init__(self):
        from textrank4zh import TextRank4Keyword, TextRank4Sentence
        self.sentence_handler = TextRank4Sentence()
        self.keywords_handler = TextRank4Keyword()

        import jieba as jieba
        self.jieba = jieba

    def get_text_from_docx(self, docx_path):
        """
        return text content of a docx file

        Parameters
        ----------
        docx_path: string
        """
        from docx import Document
        doc = Document(docx_path)
        text_list = []
        for paragraph in doc.paragraphs:
            text_list.append(paragraph.text)
        return '\n'.join(text_list)

    def get_key_words(self, input_text: str) -> list[str]:
        words = []
        for item in self.keywords_handler.get_keywords(20, word_min_len=1):
            words.append(item.word)
        return words

    def get_sentences_from_text(self, text) -> list[str]:
        """
        return a list of sentences

        Parameters
        ----------
        text: string
        """
        self.sentence_handler.analyze(text)
        return self.sentence_handler.sentences #type: ignore

    def get_key_sentences_from_text(self, text) -> list[str]:
        """
        return a list of key sentence

        Parameters
        ----------
        text: string
        """
        self.sentence_handler.analyze(text)
        sentences = [s['sentence'] for s in self.sentence_handler.get_key_sentences()]
        return sentences

    def sentence_contracting(self, text: str) -> str:
        """
        return a contracted sentence

        Parameters
        ----------
        text: string
        """
        text = text.strip()

        import jieba.posseg as pseg
        words = pseg.cut(text)
        result = []
        for word, flag in words:
            if word == "能":
                print(flag)
            if flag in ['n', 'nr', 'ns', 'v', 'r', 'p', 'r1']:
                result.append(word)
        return ''.join(result)

    def split_words_from_sentence(self, sentence):
        """
        return a list of words cut from sentence 

        Parameters
        ----------
        sentence: string
        """
        return list(self.jieba.cut(sentence))


if __name__ == "__main__":
    chinese = Chinese()
    sentence = "他穿着一件大翻领上带着道道滚边的海军服。"
    r = chinese.split_words_from_sentence(sentence)
    print(sentence)
    print(r)
    print("what i would say: 他穿着一件 大翻领上带着道道滚边的 海军服。")



import pandas as pd
import os


class Friendship():
    def __init__(self, csv_path=None):
        if csv_path == None:
            self.__df_path = os.path.join(os.getcwd(), 'friendships.csv')
        else:
            self.__df_path = os.path.abspath(csv_path)

        if not os.path.exists(self.__df_path):
            self.df = pd.DataFrame(columns=[
                'name',
                'description',
                'positive_power',
                'negative_power',
                'balanced_times'
            ])
        else:
            self.df = pd.read_csv(self.__df_path)

    def __str__(self):
        return '-' * 30 + '\n' + str(self.df.head())

    def add_person(self, name):
        temp_df = self.df.loc[self.df['name'] == name]
        if len(temp_df) >= 1:
            print('You already have it')
        else:
            self.df = self.df.append(
                {
                    'name': name,
                    'description': '',
                    'positive_power': 0,
                    'negative_power': 0,
                    'balanced_times': 0,
                },
                ignore_index=True
            )

    def delete_person(self, name):
        self.df = self.df[self.df['name'] != name]

    def giving(self, name, power):
        temp_df = self.df.loc[self.df['name'] == name]
        if len(temp_df) >= 1:
            new_value = temp_df.iloc[0]['positive_power'] + power
            temp_df.loc[:, 'positive_power'] = new_value
            temp_df = self.__calculate_balanced_times(temp_df)
            self.df.update(temp_df)
        else:
            print(f"{name} does not exist in your friendship list!")

    def taking(self, name, power):
        temp_df = self.df.loc[self.df['name'] == name]
        if len(temp_df) >= 1:
            new_value = temp_df.iloc[0]['negative_power'] + power
            temp_df.loc[:, 'negative_power'] = new_value
            temp_df = self.__calculate_balanced_times(temp_df)
            self.df.update(temp_df)
        else:
            print(f"{name} does not exist in your friendship list!")

    def __calculate_balanced_times(self, single_row_df):
        positive = single_row_df.iloc[0]['positive_power']
        negative = single_row_df.iloc[0]['negative_power']
        balanced_times = single_row_df.iloc[0]['balanced_times']
        if positive == negative:
            single_row_df.loc[:, 'balanced_times'] = balanced_times + 1
        return single_row_df

    def seek_for_help(self, description=''):
        temp_df = self.df.loc[self.df['positive_power'] > self.df['negative_power']]
        temp_df['released_power'] = temp_df['positive_power'] - temp_df['negative_power']
        sorted_df = temp_df.sort_values(['released_power', 'balanced_times'])
        print(sorted_df.head())

    def can_i_help(self, name=None):
        temp_df = self.df.loc[self.df['negative_power'] > self.df['positive_power']]
        temp_df['borrowed_power'] = temp_df['negative_power'] - temp_df['positive_power']
        sorted_df = temp_df.sort_values(['borrowed_power', 'balanced_times'])
        print(sorted_df.head())

    def commit(self):
        self.df.to_csv(self.__df_path, index=False)


if __name__ == "__main__":
    print('\n' * 50)
    friendship = Friendship('/home/yingshaoxo/Documents/friendship.csv')
    friendship.add_person('A')
    friendship.add_person('B')
    friendship.giving('A', 5)
    friendship.taking('B', 3)
    friendship.seek_for_help()
    friendship.can_i_help()
    # friendship.commit()
